# Copyright 2013, 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Classes for faking sending and receiving of mail
================================================

Copyright (C) 2006 Reahl Software Services (Pty) Ltd.  All rights reserved. (www.reahl.org)

"""
from __future__ import print_function, unicode_literals, absolute_import, division

import asyncore
import logging
from smtpd import DebuggingServer, SMTPServer
from threading import Thread
import email

from reahl.dev.devshell import WorkspaceCommand

#------------------------------------------------[ utilities ]


class EchoSMTPServer(DebuggingServer):
    def __init__(self):
        DebuggingServer.__init__(self, ('localhost', 8025), (None, 0))

    def process_message(self, peer, mailfrom, rcpttos, data):
        #Parse the message into an email.Message instance
        logging.debug("Recieved Message")
        email.message_from_string(data)
        m = email.message_from_string(data)
        logging.debug("Parsed Message")

        self.current_id = m['Message-Id']
        self.process_simple_message(m)

    def process_simple_message(self, message):
        print((self.message_as_text(message)))

    def message_as_text(self, message):
        text = "Processing Message: %s\n" % self.current_id
        text += "   From:%s \n" % message['From']
        text += "   To:%s \n" % message['To']
        text += "   Subject:%s \n" % message['Subject']
        if not message.is_multipart():
            text += "   Body:%s\n " % message.get_payload(decode=1)
        else:
            text += "   Body:\n"
            for part in message.walk():
                text += '%s \n' % part.get_payload(decode=1)
        return text


class ServeSMTP(WorkspaceCommand):
    """Runs an SMTP server on port 8025 for testing."""
    keyword = 'servesmtp'
    def execute(self, options, args):
        server = EchoSMTPServer()
        print("Running Echo SMTP Server on port 8025, type [ctrl-c] to exit.")
        try:
            asyncore.loop()
        except KeyboardInterrupt:
            pass
        

class FakeSMTPServer(EchoSMTPServer):

    def __init__(self, call_back_function=None):
        SMTPServer.__init__(self, ('localhost', 8025), (None, 0))
        self.call_back_function = call_back_function

    #A place to store the current ID of the message that we are processing.
    current_id = 0

    def handle_close(self):
        logging.debug("close called")

    def process_simple_message(self, message):
        logging.debug(self.message_as_text(message))

        if self.call_back_function:
            self.call_back_function(message)


class MailTester(object):

    def __init__(self, call_back_function=None):
        self.call_back_function = call_back_function
        self.running = False

    def main_loop(self):
        while self.running:
            logging.debug("mainloop...")
            asyncore.loop(timeout=1, count=1)
        self.fake_server.close()

    def start(self):
        self.fake_server = FakeSMTPServer(self.call_back_function)
        self.thread = Thread(target=self.main_loop)
        self.running = True
        self.thread.start()

    def stop(self):
        logging.debug("stop called")
        self.running = False
        self.thread.join()






# Copyright 2013, 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals, absolute_import, division
class EventTester(object):
    def __init__(self, event, **arguments):
        self.occurring_event = event.with_arguments(**arguments)
        self.occurring_event.from_input(self.occurring_event.as_input())

    def fire_event(self):
        return self.occurring_event.fire()

    @property
    def can_write_event(self):
        return self.occurring_event.can_write()

    @property
    def can_read_event(self):
        return self.occurring_event.can_read()



# Copyright 2013, 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module houses the main classes used to understand and manipulate Reahl projects in development."""
from __future__ import print_function, unicode_literals, absolute_import, division
import six
import os
import io
import sys
import glob
import re
import glob
import os.path
import shutil
import subprocess
import shlex
import logging
import email.utils
from contextlib import contextmanager
import datetime
import pkgutil

import babel
from pkg_resources import require, DistributionNotFound, VersionConflict, get_distribution
from setuptools import find_packages, setup
from xml.parsers.expat import ExpatError

from reahl.bzrsupport import Bzr
from reahl.component.shelltools import Executable
from reahl.dev.xmlreader import XMLReader, TagNotRegisteredException
from reahl.component.exceptions import ProgrammerError
from reahl.component.eggs import ReahlEgg
from reahl.component.py3compat import ascii_as_bytes_or_str

from reahl.dev.exceptions import NoException, StatusException, AlreadyUploadedException, NotBuiltException, NotAValidProjectException, \
    InvalidProjectFileException, NotUploadedException, NotVersionedException, NotCheckedInException, \
    MetaInformationNotAvailableException, AlreadyDebianisedException, \
    MetaInformationNotReadableException, UnchangedException, NeedsNewVersionException, \
    AlreadyMarkedAsReleasedException, NoSuchProjectException, NotBuiltAfterLastCommitException, NotBuiltException
    
class ProjectNotFound(Exception):
    pass

class InvalidXMLException(Exception):
    pass

class InvalidLocaleString(Exception):
    pass

class EggNotFound(Exception):
    pass


class DistributionPackage(object):
    """A DistributionPackage is a package that may be built for distribution."""
    def __init__(self, project):
        self.project = project
        self.repositories = []

    def inflate_attributes(self, reader, attributes, parent):
        self.__init__(parent)
        
    def inflate_child(self, reader, child, tag, parent):
        self.repositories.append(child)


class PythonSourcePackage(DistributionPackage):
    """A PythonSourcePackage is an egg as built with setup sdist."""
    @classmethod
    def get_xml_registration_info(cls):
        return ('distpackage', cls, 'sdist')

    def __str__(self):
        return 'Sdist (source egg).'

    def build(self):
        self.project.generate_setup_py()
        build_directory = os.path.join(self.project.workspace.build_directory, self.project.project_name)
        self.project.setup(['build', '-b', build_directory, 'sdist', '--dist-dir', self.project.distribution_egg_repository.root_directory])

    @property
    def is_built(self):
        return self.project.distribution_egg_repository.is_uploaded(self)

    @property
    def package_files(self):
        return [self.targz_filename]

    @property
    def targz_filename(self):
        return self.targz_filename_for(self.project)

    def targz_filename_for(self, project):
        return '%s-%s.tar.gz' % (project.project_name, project.version)

    def last_built_after(self, when):
        return self.project.distribution_egg_repository.is_uploaded_after(self, when)

    @property
    def files_to_distribute(self):
        return self.project.distribution_egg_repository.uploaded_files_for(self)

    @property
    def unique_id(self):
        return self.targz_filename


class DebianPackage(DistributionPackage):
    def __str__(self):
        return 'Debian:\t\t\t%s' % self.deb_filename(self.project)

    @classmethod
    def get_xml_registration_info(cls):
        return ('distpackage', cls, 'deb')

    @property
    def debian_build_architecture(self):
        running_command = Executable('dpkg-architecture').Popen(['-qDEB_BUILD_ARCH'], stdout=subprocess.PIPE)
        arch_string = running_command.communicate()[0].strip('\n ')
        if running_command.returncode != 0:
            raise Exception()
        return arch_string

    def deb_filename(self, egg_project):
        return 'python-%s_%s_all.deb' % (egg_project.project_name, self.project.version)
        
    @property
    def targz_filename(self):
        return 'python-%s_%s.tar.gz' % (self.project.project_name, self.project.version)

    @property
    def changes_filename(self):
        return 'python-%s_%s_%s.changes' % (self.project.project_name, self.project.version, 
                                            self.debian_build_architecture)

    @property
    def dsc_filename(self):
        return 'python-%s_%s.dsc' % (self.project.project_name, self.project.version)

    @property
    def unique_id(self):
        return self.dsc_filename

    def clean_files(self, filenames):
        for src in filenames:
            os.remove(src)

    @property
    def is_built(self):
        return self.project.distribution_apt_repository.is_uploaded(self)

    @property
    def build_output_files(self):
        return [os.path.join(self.project.directory, '..', i) 
                for i in self.package_files]

    @property
    def files_to_distribute(self):
        return self.project.distribution_apt_repository.uploaded_files_for(self)
        
    def last_built_after(self, when):
        return self.project.distribution_apt_repository.is_uploaded_after(self, when)

    def build(self):
        self.generate_install_files()
        Executable('dpkg-buildpackage').check_call(['-sa', '-rfakeroot', '-Istatic','-I.bzr', '-k%s' % os.environ['EMAIL']], cwd=self.project.directory)
        self.project.distribution_apt_repository.upload(self, [])
        self.clean_files(self.build_output_files)
        
    @property
    def package_files(self):
        return [self.deb_filename(self.project), self.targz_filename, self.changes_filename, self.dsc_filename]


    def generate_install_files(self):
        pass


class DebianDotInstallFile(object):
    def __init__(self, package_set, egg_project):
        self.egg_project = egg_project
        self.package_set = package_set

    @property
    def filename(self):
        return os.path.join('debian', 'python-%s.install' % self.egg_project.project_name)

    def generate(self):
        with io.open(self.filename, 'w') as out_file:
            out_file.write(os.path.join('debian', self.package_set.deb_base_name, self.egg_project.project_name, 'usr'))
            out_file.write('  /\n') 


class DebianPackageSet(DebianPackage):
    def __str__(self):
        binaries_files = [self.deb_filename(i) for i in self.project.egg_projects]
        return 'Debian:\t\t\t%s' % ', '.join(binaries_files)
 
    @classmethod
    def get_xml_registration_info(cls):
        return ('distpackageset', cls, 'deb')
        
    @property
    def package_files(self):
        sources_files = [self.targz_filename, self.changes_filename, self.dsc_filename]
        binaries_files = [self.deb_filename(i) for i in self.project.egg_projects]
        return sources_files + binaries_files
        
    @property
    def deb_base_name(self):
        return 'python-%s' % self.project.project_name
    
    def generate_install_files(self):
        for egg in self.project.egg_projects:
            DebianDotInstallFile(self, egg).generate()
            
            
class RepositoryLocalState(object):
    """Used by Repository objects to keep track locally of what packages have been uploaded to the Repository."""
    def __init__(self, repository):
        self.repository = repository
        self.uploaded_project_ids = set()
        
    def is_uploaded(self, package):
        return package.unique_id in self.uploaded_project_ids
        
    def set_uploaded(self, package):
        self.uploaded_project_ids.add(package.unique_id)
        
    @property
    def upload_state_filename(self):
        return os.path.join(self.repository.repository_state_directory, '%s.uploaded' % self.repository.unique_id )
        
    def read(self):
        self.uploaded_project_ids = set()
        if not os.path.exists(self.upload_state_filename):
            return
        f = io.open(self.upload_state_filename, 'r')
        self.uploaded_project_ids = set(f.read().splitlines())
        f.close()

    def write(self):
        f = io.open(self.upload_state_filename, 'w')
        f.writelines(['%s\n' % i for i in self.uploaded_project_ids])
        f.close()


class RemoteRepository(object):
    """A place where packages can be released to."""
    def __init__(self):
        self.local_storage = RepositoryLocalState(self)

    @property
    def knock_host(self):
        return '127.0.0.1'
        
    def knock(self, knocks):
        Executable('knock').check_call([self.knock_host]+knocks)

    @property
    def repository_state_directory(self):
        return self.workspace.repository_state_directory

    def is_uploaded(self, package):
        self.local_storage.read()
        return self.local_storage.is_uploaded(package)
        
    def upload(self, package, knocks):
        if not package.is_built:
            raise NotBuiltException()
        if self.is_uploaded(package):
            raise AlreadyUploadedException()
        if knocks:
            self.knock(knocks)
        self.transfer(package)
        self.local_storage.set_uploaded(package)
        self.local_storage.write()

    @property
    def unique_id(self):
        assert None, 'Not implemented'


class PackageIndex(RemoteRepository):
    """A PyPi repository."""
    @classmethod
    def get_xml_registration_info(cls):
        return ('packageindex', cls)

    def inflate_attributes(self, reader, attributes, parent):
        self.__init__(parent.project.workspace,
                      six.text_type(attributes['repository']))

    def __init__(self, workspace, repository):
        super(PackageIndex, self).__init__()
        self.workspace = workspace
        self.repository = repository

    @property
    def knock_host(self):
        return self.repository

    @property
    def unique_id(self):
        file_unsafe_id = '_'.join(['pypi', self.repository])
        return file_unsafe_id.replace(os.sep, '-')

    def transfer(self, package):
        package.project.setup(['sdist', '--dist-dir', package.project.distribution_egg_repository.root_directory, 'upload',
                               '--repository', self.repository, '-s'])

    
class SshRepository(RemoteRepository):
    """A simple Repository for uploading debian or egg packages to via ssh. Uploading merely means to scp all the necessary
       debian packaging files to a directory on the remote host."""
    @classmethod
    def get_xml_registration_info(cls):
        return ('sshdirectory', cls)

    def inflate_attributes(self, reader, attributes, parent):
        self.__init__(parent.project.workspace,
                      six.text_type(attributes['host']), 
                      six.text_type(attributes.get('login', os.environ['USER'])) or None,
                      six.text_type(attributes['destination']) or None)

    def __init__(self, workspace, host, login, destination):
        super(SshRepository, self).__init__()
        self.workspace = workspace
        self.host = host
        self.login = login
        self.destination = destination

    @property
    def knock_host(self):
        return self.host

    def transfer(self, package):
        files = package.files_to_distribute
        Executable('scp').check_call(files+['%s@%s:%s' % (self.login, self.host, self.destination)])

    @property
    def unique_id(self):
        file_unsafe_id = '_'.join([self.login, self.host, self.destination])
        return file_unsafe_id.replace(os.sep, '-')


class LocalRepository(object):
    def __init__(self, root_directory):
        self.root_directory = root_directory
        self.ensure_directory(root_directory)

    def ensure_directory(self, directory_name):
        if not os.path.isdir(directory_name):
            os.mkdir(directory_name)

    def upload(self, package, knocks):
        for filename in package.build_output_files:
            shutil.copy(filename, self.root_directory)

    def is_uploaded(self, package):
        result = True
        for filename in self.uploaded_files_for(package):
            result = result and os.path.isfile(filename)
        return result
        
    def is_uploaded_after(self, package, when):
        if not self.is_uploaded(package):
            return False
        a_file = self.uploaded_files_for(package)[0]
        return datetime.datetime.fromtimestamp(os.path.getmtime(a_file)) >= when

    def remove_uploaded(self, package):
        for filename in self.uploaded_files_for(package):
            if os.path.isfile(filename):
                os.remove(filename)

    def uploaded_files_for(self, package):
        return [os.path.join(self.root_directory, filename)
                for filename in package.package_files]

        
class LocalAptRepository(LocalRepository):
    def build_index_files(self):
        with io.open( os.path.join(self.root_directory, 'Packages'), 'w' ) as packages_file:
            Executable('apt-ftparchive').check_call(['packages', '.'], cwd=self.root_directory, stdout=packages_file)

        path_name, directory_name = os.path.split(self.root_directory)
        with io.open( os.path.join(self.root_directory, 'Release'), 'w' ) as release_file:
            Executable('apt-ftparchive').check_call(['release', directory_name], cwd=path_name, stdout=release_file)

        Executable('gpg').check_call(['-abs', '--yes', '-o', 'Release.gpg', 'Release'], cwd=self.root_directory)


class EntryPointExport(object):
    @classmethod
    def get_xml_registration_info(cls):
        return ('export', cls, None)

    def __init__(self, entrypoint, name, locator_string):
        self.entry_point = entrypoint
        self.name = name
        self.locator = EntryPointLocator(locator_string)

    def inflate_attributes(self, reader, attributes, parent):
        assert 'entrypoint' in attributes, 'No entrypoint specified'
        assert 'name' in attributes, 'No name specified'
        assert 'locator' in attributes, 'No locator specified'
        self.__init__(attributes['entrypoint'], attributes['name'], attributes['locator'])


class ScriptExport(EntryPointExport):
    @classmethod
    def get_xml_registration_info(cls):
        return ('script', cls, None)

    def __init__(self, name, locator_string):
        super(ScriptExport, self).__init__('console_scripts', name, locator_string)

    def inflate_attributes(self, reader, attributes, parent):
        assert 'name' in attributes, 'No name specified'
        assert 'locator' in attributes, 'No locator specified'
        self.__init__(attributes['name'], attributes['locator'])


class ReahlEggExport(EntryPointExport):
    def __init__(self, locator_string):
        super(ReahlEggExport, self).__init__('reahl.eggs', 'Egg', locator_string)

    def inflate_attributes(self, reader, attributes, parent):
        assert 'locator' in attributes, 'No locator specified'
        self.__init__(attributes['locator'])


class TranslationPackage(EntryPointExport):
    @classmethod
    def get_xml_registration_info(cls):
        return ('translations', cls, None)

    def __init__(self, name, locator_string):
        super(TranslationPackage, self).__init__('reahl.translations', name, locator_string)

    def inflate_attributes(self, reader, attributes, parent):
        assert 'locator' in attributes, 'No locator specified'
        assert ':' not in attributes['locator'], 'Please specify only a package name'
        self.__init__(parent.project_name, attributes['locator'])

    @property
    def path(self):
        return self.locator.package_path


class ExcludedPackage(object):
    @classmethod
    def get_xml_registration_info(cls):
        return ('excludepackage', cls, None)

    def __init__(self, name):
        self.name = name

    def inflate_attributes(self, reader, attributes, parent):
        assert 'name' in attributes, 'No name specified'
        self.__init__(attributes['name'])
    

class Version(object):
    def __init__(self, version_string):
        match = re.match('^(?P<major>\d+)\.(?P<minor>\d+)(\.(?P<patch>[0-9a-zA-Z]+)((?P<other>[^-]+)?)(-(?P<debian_rev>.*))?)?$', version_string)
        self.major = match.group('major')
        self.minor = match.group('minor')
        self.patch = match.group('patch')
        self.other = match.group('other') 
        self.debian_revision = match.group('debian_rev')

    def __str__(self):
        tail_bits = []
        if self.patch:
            tail_bits += [self.patch]
        if self.debian_revision:
            tail_bits += [self.debian_revision]

        tail = ['-'.join(tail_bits)] if tail_bits else []
        return '.'.join([self.major, (self.minor+(self.other or ''))] + tail)

    def truncated(self):
        return Version('.'.join([self.major, self.minor]))

    def is_alpha_version(self):
        if not self.patch:
            return True
        return not re.match('^\d+$', self.patch)

    def upper_version(self):
        return Version('.'.join([self.major, six.text_type(int(self.minor)+1)]))
            
    def lower_deb_version(self):
        return self.truncated()  

    def lower_egg_version(self):
        if self.is_alpha_version():
           return self.as_upstream()
        return self.truncated()

    def as_upstream(self):
        return Version('.'.join([self.major, self.minor]+([self.patch] if self.patch else [])))
        

class Dependency(object):
    def __init__(self, project, name, version=None, ignore_version=False):
        self.project = project
        self.name = name
        self.ignore_version = ignore_version
        self.version = Version(version) if version else None
        
    @classmethod
    def get_xml_registration_info(cls):
        return ('egg', cls, None)

    def inflate_attributes(self, reader, attributes, parent):
        assert 'name' in attributes, 'No name specified'
        ignore_version = attributes.get('ignoreversion', False) == 'True'
        self.__init__(parent.project, attributes['name'], attributes.get('version', None), ignore_version=ignore_version)

    def as_string_for_egg(self):
        if not self.ignore_version and self.version_is_available():
            version = self.exact_version()
            return '%s>=%s,<%s' % (self.name, version.lower_egg_version(), version.upper_version())
        return self.name

    def as_string_for_deb(self):
        if not self.ignore_version and self.version_is_available():
            version = self.exact_version()
            return 'python-%s (>=%s), python-%s (<<%s)' % (self.name, version.lower_deb_version(), 
                                                           self.name, version.upper_version())
        return 'python-%s' % self.name
    
    def as_project_dependencies(self, project_list):
        try:
            return [project_list.project_named(self.name)]
        except ProjectNotFound:
            print('WARNING: Could not find a project in the workspace for dependency named %s (ignoring)' % self.name, file=sys.stderr)
            return []

    @property
    def is_internal(self):
        return self.project.chicken_project and self.project.chicken_project.contains_project_named(self.name)

    @property
    def is_in_development(self):
        return self.project.workspace.has_project_named(self.name) 

    @property
    def is_installed(self):
        try:
            require(self.as_string_for_egg())
        except (DistributionNotFound, VersionConflict):
            return False
        return True

    def version_is_available(self):
        return self.is_internal or self.version or self.project.workspace.has_project_named(self.name) 

    def exact_version(self):
        assert self.version_is_available(), 'No version can be determined for dependency %s' % self.name
        assert not self.ignore_version, 'Versions are ignored on dependency %s' % self.name
        if self.is_internal:
            return self.project.version

        if self.version:
            return self.version

        try:
            project = self.project.workspace.project_named(self.name)
            return project.version
        except ProjectNotFound:
            return None




class ThirdpartyDependency(Dependency):
    def __init__(self, project, name, min_version=None, max_version=None):
        super(ThirdpartyDependency, self).__init__(project, name, version=min_version)
        self.max_version = Version(max_version) if max_version else None

    def inflate_attributes(self, reader, attributes, parent):
        assert 'name' in attributes, 'No name specified'
        self.__init__(parent.project, attributes['name'], min_version=attributes.get('minversion', None),
                                                          max_version=attributes.get('maxversion', None))

    @property
    def min_version(self):
        return self.version

    @classmethod
    def get_xml_registration_info(cls):
        return ('thirdpartyegg', cls, None)

    def as_string_for_deb(self):
        return ''

    def as_string_for_egg(self):
        requirement_string = self.name
        if self.min_version:
            requirement_string += '>=%s' % self.min_version
        if self.max_version:
            comma = ',' if self.min_version else ''
            requirement_string += '%s<%s' % (comma, self.max_version)
        return requirement_string


class XMLDependencyList(list):
    "Purely for reading related dependencies from XML."""
    def __init__(self, project, purpose):
        self.workspace = project.workspace
        self.project = project
        self.purpose = purpose

    @classmethod
    def get_xml_registration_info(cls):
        return ('deps', cls, None)

    def inflate_attributes(self, reader, attributes, parent):
        assert 'purpose' in attributes, 'No purpose specified'
        self.__init__(parent, attributes['purpose'])

    def inflate_child(self, reader, child, tag, parent):
        assert isinstance(child, Dependency), 'Got %s, expected a dependency' % child
        self.append(child)


class ExtrasList(list):
    "Purely for reading extras dependencies from XML."""
    def __init__(self, project, name):
        self.workspace = project.workspace
        self.project = project
        self.name = name

    @classmethod
    def get_xml_registration_info(cls):
        return ('extras', cls, None)

    def inflate_attributes(self, reader, attributes, parent):
        assert 'name' in attributes, 'No name specified'
        self.__init__(parent, attributes['name'])

    def inflate_child(self, reader, child, tag, parent):
        assert isinstance(child, Dependency), 'Got %s, expected a dependency' % child
        self.append(child)


class ConfigurationSpec(EntryPointExport):
    def __init__(self, locator_string):
        super(ConfigurationSpec, self).__init__('reahl.configspec', 'config', locator_string)

    @classmethod
    def get_xml_registration_info(cls):
        return ('configuration', cls, None)

    def inflate_attributes(self, reader, attributes, parent):
        assert 'locator' in attributes, 'No locator specified'
        self.__init__(attributes['locator'])


class ScheduledJobSpec(EntryPointExport):
    def __init__(self, locator_string):
        super(ScheduledJobSpec, self).__init__('reahl.scheduled_jobs', None, locator_string)
        self.name = self.locator.class_name

    @classmethod
    def get_xml_registration_info(cls):
        return ('schedule', cls, None)

    def inflate_attributes(self, reader, attributes, parent):
        assert 'locator' in attributes, 'No locator specified'
        self.__init__(attributes['locator'])


class OrderedPersistedClass(object):
    def __init__(self, locator_string, order, entry_point):
        self.entry_point = entry_point
        self.locator = EntryPointLocator(locator_string)
        self.order = order

    @property
    def name(self):
        return six.text_type(self.order)
    
    @classmethod
    def get_xml_registration_info(cls):
        return ('class', cls, None)

    def inflate_attributes(self, reader, attributes, parent):
        assert 'locator' in attributes, 'No locator specified'
        self.__init__(attributes['locator'], len(parent), parent.entry_point)


class OrderedClassesList(list):
    @property
    def entry_point(self):
        raise ProgrammerError('subclasses must provide this')
    
    def inflate_child(self, reader, child, tag, parent):
        assert isinstance(child, OrderedPersistedClass), 'Got %s, expected a class' % child
        self.append(child)


class PersistedClassesList(OrderedClassesList):
    "Purely for reading a list of persisted classes XML."""
    @property
    def entry_point(self):
        return 'reahl.persistlist'

    @classmethod
    def get_xml_registration_info(cls):
        return ('persisted', cls, None)


class MigrationList(OrderedClassesList):
    "Purely for reading a list of migration classes XML."""
    @property
    def entry_point(self):
        return 'reahl.migratelist'

    @classmethod
    def get_xml_registration_info(cls):
        return ('migrations', cls, None)


class FileList(list):
    @property
    def entry_point(self):
        return 'reahl.attachments.any'

    @classmethod
    def get_xml_registration_info(cls):
        return ('staticfiles', cls, None)

    def inflate_child(self, reader, child, tag, parent):
        assert isinstance(child, ShippedFile), 'Got %s, expected a StaticFile' % child
        self.append(child)


class AttachmentList(list):
    def __init__(self, file_type):
        super(AttachmentList, self).__init__()
        self.file_type = file_type
    
    @classmethod
    def get_xml_registration_info(cls):
        return ('attachedfiles', cls, None)

    @property
    def entry_point(self):
        return 'reahl.attachments.%s' % self.file_type
    
    def inflate_attributes(self, reader, attributes, parent):
        assert 'filetype' in attributes, 'No filetype specified'
        file_type = attributes['filetype']
        assert file_type in ['js', 'css'], 'filetype %s undefined. Please use one of "js" or "css"' % file_type
        self.__init__(file_type)

    def inflate_child(self, reader, child, tag, parent):
        assert isinstance(child, ShippedFile), 'Got %s, expected a class' % child
        self.append(child)


class ShippedFile(object):
    def __init__(self, path, order, entry_point):
        self.entry_point = entry_point
        self.path = path
        self.order = order

    @property
    def name(self):
        return '%s:%s' % (self.order, self.path)
        
    @property
    def locator(self):
        return EntryPointLocator('reahl')
    
    @classmethod
    def get_xml_registration_info(cls):
        return ('file', cls, None)

    def inflate_attributes(self, reader, attributes, parent):
        assert 'path' in attributes, 'No path specified'
        self.__init__(attributes['path'], len(parent), parent.entry_point)


class EntryPointLocator(object):
    def __init__(self, string_spec):
        self.string_spec = string_spec
    @property
    def package_name(self):
        return self.string_spec.split(':')[0]
    @property
    def package_path(self):
        return self.package_name.replace('.', os.path.sep)
    @property
    def class_name(self):
        return self.string_spec.split(':')[-1]



class NamespaceEntry(object):
    def __init__(self, name):
        self.name = name

    @classmethod
    def get_xml_registration_info(cls):
        return ('package', cls, None)

    def inflate_attributes(self, reader, attributes, parent):
        assert 'name' in attributes, 'No name specified'
        self.__init__(attributes['name'])

    @property
    def path(self):
        return EntryPointLocator(self.name).package_path

    
class NamespaceList(list):
    @classmethod
    def get_xml_registration_info(cls):
        return ('namespaces', cls, None)

    def inflate_child(self, reader, child, tag, parent):
        assert isinstance(child, NamespaceEntry), 'Got %s, expected a NamespaceEntry' % child
        self.append(child)


class NoBasket(object):
    def __init__(self, workspace):
        self.workspace = workspace
        self.project_name = 'NO BASKET'



class CommandAlias(object):
    @classmethod
    def get_xml_registration_info(cls):
        return ('alias', cls, None)

    def __init__(self, name, full_command):
        self.full_command = full_command
        self.name = name

    def inflate_attributes(self, reader, attributes, parent):
        assert 'command' in attributes, 'No command specified'
        assert 'name' in attributes, 'No arguments specified'
        self.__init__(attributes['name'], attributes['command'])
    
    @property
    def arg_list(self):
        return shlex.split(self.full_command)[1:]

    @property
    def command(self):
        return shlex.split(self.full_command)[0]


class ExtraPath(object):
    @classmethod
    def get_xml_registration_info(cls):
        return ('pythonpath', cls, None)

    def __init__(self, path):
        self.path = path

    def inflate_attributes(self, reader, attributes, parent):
        assert 'path' in attributes, 'No path specified'
        self.__init__(attributes['path'])



class ProjectTag(object):
    @classmethod
    def get_xml_registration_info(cls):
        return ('tag', cls, None)

    def __init__(self, name):
        self.name = name

    def inflate_attributes(self, reader, attributes, parent):
        assert 'name' in attributes, 'No name specified'
        self.__init__(attributes['name'])
    


class ProjectMetadata(object):
    def __str__(self):
        return 'Default project metadata provider'
    
    def __init__(self, project):
        self.project = project
        
    @property
    def project_name(self):
        return os.path.basename(self.directory)

    @property
    def directory(self):
        return self.project.directory

    @property
    def version(self):
        if self.project.chicken_project:
            return self.project.chicken_project.version
        return Version('0.0')

    def info_readable(self):
        if self.project.chicken_project:
            return self.project.chicken_project.info_readable()
        return True
    
    def info_completed(self):
        if self.project.chicken_project:
            return self.project.chicken_project.info_completed()
        return True

    def get_url_for(self, project):
        if self.project.chicken_project:
            return self.project.chicken_project.get_url_for(project)
        return 'No url provided'

    def get_description_for(self, project):
        if self.project.chicken_project:
            return self.project.chicken_project.get_description_for(project)
        return 'No description provided'
        
    def get_long_description_for(self, project):
        if self.project.chicken_project:
            return self.project.chicken_project.get_long_description_for(project)
        return 'No description provided'

    @property
    def maintainer_name(self):
        if self.project.chicken_project:
            return self.project.chicken_project.maintainer_name
        return 'No maintainer provided'

    @property
    def maintainer_email(self):
        if self.project.chicken_project:
            return self.project.chicken_project.maintainer_email
        return 'No maintainer provided'



class MetaInfo(object):
    @classmethod
    def get_xml_registration_info(cls):
        return ('info', cls, None)

    def __init__(self, name):
        self.name = name

    def inflate_attributes(self, reader, attributes, parent):
        assert 'name' in attributes, 'No name specified'
        self.__init__(attributes['name'])

    def inflate_text(self, reader, text, parent):
        self.contents = text


class HardcodedMetadata(ProjectMetadata):
    def __init__(self, parent):
        super(HardcodedMetadata, self).__init__(parent)
        self.info = {}
        
    def __str__(self):
        return 'Metadata read from a .reahlproject file'

    @classmethod
    def get_xml_registration_info(cls):
        return ('metadata', cls, 'reahlproject')

    def inflate_attributes(self, reader, attributes, parent):
        self.__init__(parent)

    def inflate_child(self, reader, child, tag, parent):
        assert isinstance(child, MetaInfo), 'Got %s, expected an info' % child
        self.info[child.name] = child
            
    @property
    def version(self):
        return Version(self.info['version'].contents)

    def get_long_description_for(self, project):
        return self.info['long_description'].contents

    def get_description_for(self, project):
        return self.info['description'].contents
    
    def get_url_for(self, project):
        return self.info['url'].contents

    @property
    def maintainer_name(self):
        return self.info['maintainer_name'].contents

    @property
    def maintainer_email(self):
        return self.info['maintainer_email'].contents

    def info_readable(self):
        return self.info_completed()

    def info_completed(self):
        expected_info = {'url', 'version', 'description', 'long_description', 'maintainer_name', 'maintainer_email'}
        completed_info = set(self.info.keys())
        return expected_info == completed_info



class DebianPackageMetadata(ProjectMetadata):
    def __init__(self, parent, url=None):
        super(DebianPackageMetadata, self).__init__(parent)
        self.url = url

    def __str__(self):
        return 'Debian package metadata provider'

    @classmethod
    def get_xml_registration_info(cls):
        return ('metadata', cls, 'debian')

    def inflate_attributes(self, reader, attributes, parent):
        assert 'url' in attributes.keys(), 'No url specified for project in %s' % parent.directory
        self.__init__(parent, attributes['url'])
        self.debian_control = DebianControl(os.path.join(self.project.directory, 'debian', 'control'))

    @property
    def changelog(self):
        try:
            filename = os.path.join(self.directory, 'debian', 'changelog')
            return DebianChangelog(filename)
        except IOError:
            return None

    @property
    def project_name(self):
        if self.changelog:
            deb_name = self.changelog.package_name
            assert deb_name.startswith('python-')
            return deb_name[len('python-'):]
        else:
            logging.warn('Project not debianised, using directory name for project name')
            print(':::%s' % self.directory)
            return os.path.basename(self.directory)
        
    @property
    def version(self):
        return Version(six.text_type(self.changelog.version))

    def get_long_description_for(self, project):
        return self.debian_control.get_long_description_for('python-%s' % project.project_name).replace(' . ', '\n\n')

    def get_description_for(self, project):
        return self.debian_control.get_short_description_for('python-%s' % project.project_name)
    
    def get_url_for(self, project):
        return self.url

    @property
    def maintainer_name(self):
        return self.debian_control.maintainer_name

    @property
    def maintainer_email(self):
        return self.debian_control.maintainer_email

    @property
    def deb_package_name(self):
        return 'python-%s' % self.project_name

    def generate_deb_substvars(self):
        to_do = [self.project]
        if self.project.has_children:
            to_do = self.project.egg_projects

        for current_project in to_do:
            current_vars = SubstvarsFile('debian/python-%s.substvars' % current_project.project_name)
            current_vars.read()
            current_vars['reahl:Depends'] = ', '.join([i.as_string_for_deb() for i in current_project.run_deps
                                                       if i.as_string_for_deb()])
            current_vars.write()

    def info_readable(self):
        return self.info_completed()

    def info_completed(self):
        return self.changelog is not None

    def debianise(self):
        if self.info_completed():
            raise AlreadyDebianisedException()
        
        import tempfile
        deb_dir = os.path.join(tempfile.gettempdir(), 'python-%s-0.1.0' % self.project.project_name)
        if os.path.isdir(deb_dir):
            shutil.rmtree(deb_dir)
        os.mkdir(deb_dir)
    
        dh_make_params = ['--native --defaultless --single --copyright GPL']
        if self.project.workspace.dh_make_directory:
            dh_make_params[0] += ' --templates %s' % self.project.workspace.dh_make_directory

        Executable('dh_make').check_call(dh_make_params.split(' '), cwd=deb_dir, shell=True)
            
        shutil.move(os.path.join(deb_dir, 'debian'), os.path.join(self.project.directory, 'debian'))
        shutil.rmtree(deb_dir)
        return 0


class DebianChangelog(object):
    package_name_regex = '(?P<package_name>[a-z][a-z0-9\-]*)'
    version_regex = '\((?P<version>[a-zA-Z\.0-9\-]+)\)'
    heading_regex = '^%s\s+%s.*$' % (package_name_regex, version_regex)

    def __init__(self, filename):
        self.filename = filename

    def parse_heading_for(self, element):
        with io.open(self.filename) as changelog_file:
            for line in changelog_file:
                if line.strip():
                    match = re.match(self.heading_regex, line)
                    assert match, 'Cannot parse changelog file: %s' % self.filename
                    return match.group(element)
        
    @property
    def package_name(self):
        return self.parse_heading_for('package_name')

    @property
    def version(self):
        return self.parse_heading_for('version')


class DebianControl(object):
    def __init__(self, filename):
        self.filename = filename

    @property
    def stanzas(self):
        stanzas = []
        current_stanza = ''
        with io.open(self.filename) as control_file:
            for line in control_file:
                if not line.strip():
                    if current_stanza:
                        stanzas.append(email.message_from_string(current_stanza))
                        current_stanza = ''
                else:
                    current_stanza += line
        return stanzas

    @property
    def source_stanza(self):
        [source] = [stanza for stanza in self.stanzas if 'Source' in stanza]
        return source

    def get_package_stanza(self, package_name):
        try:
            [stanza] = [stanza for stanza in self.stanzas
                        if stanza.get('Package', None) == package_name]
            return stanza
        except ValueError:
            raise AssertionError('Could not find a stanza for a package named %s in debian control file %s' % \
                                 (package_name, self.filename))
    
    @property
    def maintainer_name(self):
        return email.utils.parseaddr(self.source_stanza['Maintainer'])[0]

    @property
    def maintainer_email(self):
        return email.utils.parseaddr(self.source_stanza['Maintainer'])[1]
        
    def get_long_description_for(self, package_name):
        description_field = self.get_package_stanza(package_name)['Description']
        return ''.join([line.strip()+' ' for line in description_field.split('\n')[1:]])

    def get_short_description_for(self, package_name):
        description_field = self.get_package_stanza(package_name)['Description']
        return description_field.split('\n')[0]


        
class SourceControlSystem(object):
    def __str__(self):
        if self.project.chicken_project:
            return six.text_type(self.project.chicken_project.source_control)
        return 'No source control system selected'

    def __init__(self, project):
        self.project = project

    @property
    def last_commit_time(self):
        if self.project.chicken_project:
            return self.project.chicken_project.source_control.last_commit_time
        return datetime.datetime.fromtimestamp(0)

    def is_unchanged(self):
        if self.project.chicken_project:
            return self.project.chicken_project.source_control.is_unchanged()
        return False

    def needs_new_version(self):
        if self.project.chicken_project:
            return self.project.chicken_project.source_control.needs_new_version()
        return True

    def is_version_controlled(self):
        if self.project.chicken_project:
            return self.project.chicken_project.source_control.is_version_controlled()
        return False
    
    def is_checked_in(self):
        if self.project.chicken_project:
            return self.project.chicken_project.source_control.is_checked_in()
        return False

    def place_tag(self, tag):
        assert not self.project.chicken_project, 'You cannot tag a subproject'
        return False



class BzrSourceControl(SourceControlSystem):
    def __str__(self):
        return 'Bzr source control'

    @classmethod
    def get_xml_registration_info(cls):
        return ('sourcecontrol', cls, 'bzr')

    def inflate_attributes(self, reader, attributes, parent):
        self.__init__(parent)

    def __init__(self, parent):
        super(BzrSourceControl, self).__init__(parent)
        self.bzr = Bzr(self.project.directory)
        
    @property
    def last_commit_time(self):
        return self.bzr.last_commit_time
    
    def is_unchanged(self):
        tag = six.text_type(self.project.version)
        return tag in self.bzr.get_tags(head_only=True)

    def needs_new_version(self):
        tag = six.text_type(self.project.version)
        return tag in self.bzr.get_tags()

    def is_version_controlled(self):
        return self.bzr.is_version_controlled()

    def is_checked_in(self):
        return self.bzr.is_checked_in()

    def place_tag(self, tag):
        self.bzr.tag(tag)


class Project(object):
    """Instances of Project each represent a Reahl project in a development environment.
    """
    has_children = False
    @classmethod
    def get_xml_registration_info(cls):
        return ('project', cls, None)

    @classmethod
    def from_file(cls, workspace, directory):
        project_filename = os.path.join(directory, '.reahlproject')
        if not os.path.isfile(project_filename):
            raise NotAValidProjectException(project_filename)
        input_file = io.open(project_filename, 'r')
        try:
            reader = XMLReader()
            project = reader.read_file(input_file, (workspace, directory))
        except (TagNotRegisteredException, ExpatError, InvalidXMLException) as ex:
            raise InvalidProjectFileException(project_filename, ex)
        finally:
            input_file.close()
        return project

    def __init__(self, workspace, directory):
        self.workspace = workspace
        self.directory = directory
        self.relative_directory = os.path.relpath(directory, self.workspace.directory)

        self.packages = []
        self.python_path = []
        self.command_aliases = {}
        self._tags = []
        self.metadata = ProjectMetadata(self)
        self.source_control = SourceControlSystem(self)

    def inflate_attributes(self, reader, attributes, parent):
        workspace, directory = parent
        self.__init__(workspace, directory)

    def inflate_child(self, reader, child, tag, parent):
        if isinstance(child, CommandAlias):
            self.command_aliases[child.name] = [child.command]+child.arg_list
        elif isinstance(child, ProjectTag):
            self._tags.append(child.name)
        elif isinstance(child, ProjectMetadata):
            self.metadata = child
        elif isinstance(child, SourceControlSystem):
            self.source_control = child
        elif isinstance(child, ExtraPath):
            self.python_path.append(child.path)
        else:
            raise InvalidXMLException('Unexpected type read from XML: %s' % child)

    @property
    def tags(self):
        return self._tags

    @property
    def distribution_apt_repository(self):
        return self.workspace.distribution_apt_repository

    @property
    def distribution_egg_repository(self):
        return self.workspace.distribution_egg_repository

    @property
    def project_filename(self):
        return os.path.join(self.directory, '.reahlproject')

    @property
    def packages_to_distribute(self):
        return self.packages

    @contextmanager
    def paths_set(self):
        cwd = os.getcwd()
        path = sys.path[:]
        sys.path.extend(self.python_path)
        try:
            os.chdir(self.directory)
            yield
        finally:
            sys.path[:] = path
            os.chdir(cwd)
        
    def command_for_alias(self, alias):
        return self.command_aliases.get(alias, None)

    def list_missing_dependencies(self, for_development=None):
        pass
    #-----------[ metadata stuff ]---------------------------------------

    def info_completed(self):
        return self.metadata.info_completed()
    
    def info_readable(self):
        return self.metadata.info_readable()
        
    @property
    def project_name(self):
        return self.metadata.project_name

    @property
    def project_name_pythonised(self):
        return self.project_name.replace('-', '_')
    
    @property
    def version(self):
        return self.metadata.version

    def get_url_for(self, project):
        return self.metadata.get_url_for(project)
        
    def get_description_for(self, project):
        return self.metadata.get_description_for(project)
        
    def get_long_description_for(self, project):
        return self.metadata.get_long_description_for(project)

    @property
    def maintainer_name(self):
        return self.metadata.maintainer_name

    @property
    def maintainer_email(self):
        return self.metadata.maintainer_email

    def do_release_checks(self):
        if not self.is_version_controlled():
            raise NotVersionedException(self)
        elif not self.is_checked_in():
            raise NotCheckedInException(self)
        elif not self.info_completed():
            raise MetaInformationNotAvailableException(self)
        elif not self.info_readable():
            raise MetaInformationNotReadableException(self)

        if self.is_unchanged():
            raise UnchangedException(self)
        elif self.needs_new_version():
            raise NeedsNewVersionException(self)

        if not self.is_built():
            raise NotBuiltException()
        elif not self.latest_build_is_up_to_date_with_source():
            raise NotBuiltAfterLastCommitException()

    @property
    def status(self):
        try:
            self.do_release_checks()
            self.check_uploaded()
        except StatusException as ex:
            return ex.legend

        return NoException.legend

    def is_unchanged(self):
        return self.source_control.is_unchanged()

    def needs_new_version(self):
        return self.source_control.needs_new_version()

    def is_version_controlled(self):
        return self.source_control.is_version_controlled()

    def is_checked_in(self):
        return self.source_control.is_checked_in()

    def latest_build_is_up_to_date_with_source(self):
        last_commit_time = self.source_control.last_commit_time

        is_up_to_date = True
        for i in self.packages_to_distribute:
            is_up_to_date &= i.last_built_after(last_commit_time)
        return is_up_to_date
        
    def build(self):
        assert self.packages_to_distribute, 'For %s: No <package>... listed in .reahlproject, nothing to do.' % self.project_name
        for i in self.packages_to_distribute:
            i.build()

    def is_built(self):
        is_built = True
        for i in self.packages_to_distribute:
            is_built &= i.is_built
        return is_built
        
    def upload(self, knocks=[]):
        self.do_release_checks()
        for i in self.packages_to_distribute:
            for repo in i.repositories:
                repo.upload(i, knocks)

    def check_uploaded(self):
        for i in self.packages_to_distribute:
            for repo in i.repositories:
                if not repo.is_uploaded(i):
                    raise NotUploadedException(i)

    def mark_as_released(self):
        self.do_release_checks()
        self.check_uploaded()
        self.source_control.place_tag(six.text_type(self.version))


class EggProject(Project):
    basket_name = None
    def __init__(self, workspace, directory, include_package_data):
        super(EggProject, self).__init__(workspace, directory)
        self.namespaces = []
        self.explicitly_specified_entry_points = []
        self.excluded_packages = []
        self.include_package_data = include_package_data

        self.extras_required = {}
        self.run_deps = []
        self.build_deps = []
        self.test_deps = []

        self.translation_package = None

        self.persist_list = []
        self.migration_list = []
        self.scheduled_jobs = []
        
        self.static_files = []
        self.js_attach_list = []
        self.css_attach_list = []

    @property
    def entry_points(self):
        added_entry_points = [ReahlEggExport('reahl.component.eggs:ReahlEgg')]
        return self.migration_list + \
               self.persist_list + \
               self.static_files + \
               self.js_attach_list + \
               self.css_attach_list + \
               self.explicitly_specified_entry_points + \
               ([self.translation_package] if self.translation_package else []) +\
               added_entry_points
    
    @property
    def tags(self):
        extras = ['component']
        if not self.chicken_project:
            extras.append('toplevel')
        return super(EggProject, self).tags + extras

    @property
    def chicken_project(self):
        parent_directory = os.path.dirname(self.directory)
        if not os.path.isfile(os.path.join(parent_directory, '.reahlproject')):
            return None

        parent_project = self.get_project_in_directory(parent_directory)
        return parent_project if parent_project.has_children else None

    def get_project_in_directory(self, directory):
        try:
            return self.workspace.project_in(directory)
        except ProjectNotFound:
            return Project.from_file(self.workspace, directory)
        
    @classmethod
    def get_xml_registration_info(cls):
        return ('project', cls, 'egg')

    def inflate_attributes(self, reader, attributes, parent):
        workspace, directory = parent
        include_package_data = False
        if attributes.get('packagedata', None) == 'included':
            include_package_data = True
        if 'basket' in attributes:
            self.basket_name = attributes['basket']
        self.__init__(workspace, directory, include_package_data)

    def inflate_child(self, reader, child, tag, parent):
        if isinstance(child, DistributionPackage):
            self.packages.append(child)
        elif isinstance(child, ExtrasList):
            self.extras_required[child.name] = child
        elif isinstance(child, XMLDependencyList):
            list_to_use = getattr(self, '%s_deps' % child.purpose)
            list_to_use.extend(child)
        elif isinstance(child, NamespaceList):
            self.namespaces = child
        elif isinstance(child, TranslationPackage):
            self.translation_package = child
        elif isinstance(child, EntryPointExport):
            self.explicitly_specified_entry_points.append(child)
        elif isinstance(child, PersistedClassesList):
            self.persist_list = child
        elif isinstance(child, MigrationList):
            self.migration_list = child
        elif isinstance(child, AttachmentList):
            if child.file_type == 'js':
                self.js_attach_list = child
            else:
                self.css_attach_list = child
        elif isinstance(child, FileList):
            self.static_files = child
        elif isinstance(child, ExcludedPackage):
            self.excluded_packages.append(child)
        elif isinstance(child, ScheduledJobSpec):
            self.scheduled_jobs.append(child)
        else:
            super(EggProject, self).inflate_child(reader, child, tag, parent)

    @property
    def basket(self):
        baskets = [i for i in self.workspace.projects
                   if i.contains(self)]
        if not baskets:
            return NoBasket(self.workspace)
        else:
            # If it breaks here, it means there is more than one basket found in the current workspace which claim
            #  that this egg belongs to them.
            # Eggs are not allowed to belong to more than one basket, but this could happen if multiple versions
            #  of an egg is checked out in the current workspace
            [basket] = baskets
            return basket

    def contains(self, other):
        return False

    def get_build_sources(self):
        self.setup(['egg_info'])

        sources = []
        sources_filename = os.path.join(self.egg_info_directory, 'SOURCES.txt')

        with io.open(sources_filename) as in_file:
            for line in in_file:
                sources.append(line)
        return sources
                
    @property
    def egg_info_directory(self):
        return os.path.join(self.directory, '%s.egg-info' % self.project_name_pythonised)

    def list_missing_dependencies(self, for_development=False):
        deps = self.run_deps
        if for_development:
            deps = self.run_deps + self.build_deps + self.test_deps
        dependencies = [ i for i in deps
                         if (not i.is_in_development) and (not i.is_installed) ]
        return [i.as_string_for_egg().replace(' ', '') for i in dependencies]

    def setup(self, setup_command):
        with self.paths_set():
            return setup(script_args=setup_command,
                     name=ascii_as_bytes_or_str(self.project_name),
                     version=self.version_for_setup(),
                     description=self.get_description_for(self),
                     long_description=self.get_long_description_for(self),
                     url=self.get_url_for(self),
                     maintainer=self.maintainer_name,
                     maintainer_email=self.maintainer_email,
                     packages=self.packages_for_setup(),
                     py_modules=self.py_modules_for_setup(),
                     include_package_data=self.include_package_data,
                     package_data=self.package_data_for_setup(),
                     namespace_packages=self.namespace_packages_for_setup(),
                     install_requires=self.run_deps_for_setup(),
                     setup_requires=self.build_deps_for_setup(),
                     tests_require=self.test_deps_for_setup(),
                     test_suite=self.test_suite_for_setup(),
                     entry_points=self.entry_points_for_setup(),
                     extras_require=self.extras_require_for_setup() )

    def generate_setup_py(self):
        with io.open(os.path.join(self.directory, 'setup.py'), 'w') as setup_file:
            setup_file.write('from setuptools import setup\n')
            setup_file.write('setup(\n')
            setup_file.write('    name=%s,\n' % repr(self.project_name))
            setup_file.write('    version=%s,\n' % repr(self.version_for_setup()))
            setup_file.write('    description=%s,\n' % repr(self.get_description_for(self)))
            setup_file.write('    long_description=%s,\n' % repr(self.get_long_description_for(self)))
            setup_file.write('    url=%s,\n' % repr(self.get_url_for(self))),
            setup_file.write('    maintainer=%s,\n' % repr(self.maintainer_name))
            setup_file.write('    maintainer_email=%s,\n' % repr(self.maintainer_email))
            setup_file.write('    packages=%s,\n' % repr(self.packages_for_setup()))
            setup_file.write('    py_modules=%s,\n' % repr(self.py_modules_for_setup()))
            setup_file.write('    include_package_data=%s,\n' % repr(self.include_package_data))
            setup_file.write('    package_data=%s,\n' % repr(self.package_data_for_setup()))
            setup_file.write('    namespace_packages=%s,\n' % repr(self.namespace_packages_for_setup()))
            setup_file.write('    install_requires=%s,\n' % repr(self.run_deps_for_setup()))
            setup_file.write('    setup_requires=%s,\n' % repr(self.build_deps_for_setup()))
            setup_file.write('    tests_require=%s,\n' % repr(self.test_deps_for_setup()))
            setup_file.write('    test_suite=%s,\n' % repr(self.test_suite_for_setup()))

            setup_file.write('    entry_points={\n')
            entry_points = self.entry_points_for_setup()
            for name, values in entry_points.items():
                setup_file.write('        %s: [\n            ' % repr(name))
                value_string = ',\n            '.join([repr(value) for value in values])
                setup_file.write(value_string)
                setup_file.write('    ],\n')
            setup_file.write('                 },\n')

            setup_file.write('    extras_require=%s\n' % repr(self.extras_require_for_setup()) )
            setup_file.write(')\n')

    def version_for_setup(self):
        return ascii_as_bytes_or_str(six.text_type(self.version.as_upstream()))
        
    def run_deps_for_setup(self):
        return [ascii_as_bytes_or_str(dep.as_string_for_egg()) for dep in self.run_deps]

    def build_deps_for_setup(self):
        return [ascii_as_bytes_or_str(dep.as_string_for_egg()) for dep in self.build_deps]

    def test_deps_for_setup(self):
        return [ascii_as_bytes_or_str(dep.as_string_for_egg()) for dep in self.test_deps]

    def packages_for_setup(self):
        exclusions = [i.name for i in self.excluded_packages]
        exclusions += ['%s.*' % i.name for i in self.excluded_packages]
        return [ascii_as_bytes_or_str(i) for i in find_packages(where=self.directory, exclude=exclusions)]

    def namespace_packages_for_setup(self):
        return [ascii_as_bytes_or_str(i.name) for i in self.namespaces]  # Note: this has to return non-six.text_type strings for setuptools!

    def py_modules_for_setup(self):
        return list(set(['setup']+[i[1] for i in pkgutil.iter_modules('.') if not i[2]]))

    def package_data_for_setup(self):
        return {ascii_as_bytes_or_str(''): [ascii_as_bytes_or_str('*/LC_MESSAGES/*.mo')]}

    @property
    def test_suite(self):
        directories_to_search = [self.directory]+[os.path.join(self.directory,i.path) for i in self.namespaces]

        for directory in directories_to_search:
            for name in os.listdir(directory):
                full_path = os.path.join(directory,name)
                if os.path.isdir(full_path) and name.endswith('_dev'):
                    path = full_path[len(self.directory)+1:]
                    return '.'.join(path.split(os.path.sep))

        return 'tests'

    def test_suite_for_setup(self):
        return self.test_suite
        
    def entry_points_for_setup(self):
        entry_point_dict = {}
        for entry_point_name, entry_point in [(i.entry_point, i) for i in self.entry_points]:
            exports = entry_point_dict.setdefault(entry_point_name, [])
            exports.append('%s = %s' % (entry_point.name, entry_point.locator.string_spec))
        return entry_point_dict

    def extras_require_for_setup(self):
        return dict( [ (name, [dep.as_string_for_egg() for dep in dependencies]) 
                       for name, dependencies in self.extras_required.items()] )

    def debinstall(self, args):
        root = os.path.join(self.directory, 'debian', 'python-%s' % self.project_name)
        executable = '/usr/bin/python'
        build_directory = os.path.join(self.workspace.build_directory, self.project_name)
        self.setup(['build', '-b', build_directory, 'build_scripts', '-e', executable, 'install', '--single-version-externally-managed', '--root=%s' % root, '--prefix=/usr'])

    def distributed_python_files(self):
        package_dirs = [i.replace('.', os.sep) for i in self.packages_for_setup()]
        py_files = set()
        for package_dir in package_dirs:
            for dirname, dirnames, filenames in os.walk(package_dir):
                for filename in filenames:
                    if filename.endswith('.py'):
                        py_files.add(os.path.join(dirname, filename))
        return list(py_files)

    @property
    def locale_dirname(self):
        assert self.translation_package, 'No <translations.../> tag specified for project: "%s"' % (self.project_name)
        return os.path.join(self.directory, self.translation_package.path)

    @property
    def locale_domain(self):
        return self.project_name

    @property
    def pot_filename(self):
        return os.path.join(self.locale_dirname, self.locale_domain)
        
    def extract_messages(self, args):
        if self.translation_package:
            self.setup(['extract_messages', 
                        '--input-dirs', '.',
                        '--output-file', self.pot_filename])
        else:
            logging.warn('No <translations.../> tag specified for project: "%s"' % (self.project_name)) 

    @property
    def translated_domains(self):
        if self.translation_package:
            filenames = glob.glob(os.path.join(self.locale_dirname, '*/LC_MESSAGES/*.po'))
            return {os.path.splitext(os.path.basename(i))[0] for i in filenames}
        else:
            logging.warn('No <translations.../> tag specified for project: "%s"' % (self.project_name)) 
            return []
        
    def merge_translations(self):
        for source_dist_spec in self.translated_domains:
            try:
                source_egg = ReahlEgg(get_distribution(source_dist_spec)) 
            except DistributionNotFound:
                raise EggNotFound(source_dist_spec)
            if not os.path.isdir(self.locale_dirname):
                os.mkdir(self.locale_dirname)
            self.setup(['update_catalog', 
                        '--input-file', source_egg.translation_pot_filename,
                        '--domain', source_egg.name, 
                        '-d', self.locale_dirname])

    def compile_translations(self):
        for domain in self.translated_domains:
            self.setup(['compile_catalog', 
                        '--domain', domain, 
                        '-d', self.locale_dirname])

    def add_locale(self, locale, source_dist_spec):
        try:
            babel.parse_locale(locale)
        except ValueError:
            raise InvalidLocaleString(locale)
        try:
            source_egg = ReahlEgg(get_distribution(source_dist_spec)) 
        except DistributionNotFound:
            raise EggNotFound(source_dist_spec)
        self.setup(['init_catalog', 
                    '--input-file', source_egg.translation_pot_filename, 
                    '--domain', source_egg.name, 
                    '-d', self.locale_dirname, 
                    '--locale', locale])


class ChickenProject(EggProject):
    has_children = True
    @classmethod
    def get_xml_registration_info(cls):
        return ('project', cls, 'chicken')

    def inflate_child(self, reader, child, tag, parent):
        if isinstance(child, DebianPackageSet):
            self.packages_to_distribute.append(child)
        else:
            super(ChickenProject, self).inflate_child(reader, child, tag, parent)

    @property
    def egg_project_names(self):
        return [os.path.basename(os.path.dirname(i)) 
                for i in glob.glob(os.path.join(self.directory, '*', '.reahlproject'))]

    @property
    def egg_projects(self):
        return [self.workspace.project_named(i) for i in self.egg_project_names]

    def debinstall(self, args):
        for egg in self.egg_projects:
            root = os.path.join(self.directory,'debian', 'python-%s' % self.project_name, egg.project_name)
            executable = '/usr/bin/python'
            build_directory = os.path.join(self.workspace.build_directory, self.project_name, egg.project_name)
            egg.setup(['build', '-b', build_directory, 'build_scripts', '-e', executable, 'install', '--single-version-externally-managed', '--root=%s' % root, '--prefix=/usr'])

    def contains_project_named(self, name):
        return name in self.egg_project_names

    @property
    def tags(self):
        return super(ChickenProject, self).tags + ['toplevel']


    def packages_for_setup(self):
        return []

    def namespace_packages_for_setup(self):
        return []



class DirectoryList(list):
    """Utility class used to read a list of directory names from a file on disk."""

    @classmethod
    def from_file(cls, filename):
        directories = DirectoryList()
        directories.read(filename)
        return directories

    def read(self, filename):
        f = io.open(filename, 'r')
        dirnames = f.read().splitlines()
        f.close()
        for name in dirnames:
            self.append(name)


class ProjectList(list):
    """A list of Projects, allowing a number of operations on such a list. These include saving the list to disk."""

    def __init__(self, workspace):
        super(ProjectList, self).__init__()
        self.workspace = workspace
        self.name_index = {}  # The dict is used to speed up queries (see project_named)

    def append(self, something, ignore_duplicates=False):
        assert ignore_duplicates or something.project_name not in self.name_index, 'Attempt to add duplicate project to project list'
        if something.project_name not in self.name_index:
            self.name_index[something.project_name] = something
            super(ProjectList, self).append(something)

    def project_in(self, directory):
        full_path = os.path.normpath(directory)
        for project in self:
            if project.directory == full_path:
                return project
        raise ProjectNotFound(full_path)
    
    def has_project_named(self, name):
        return name in self.name_index.keys()

    def project_named(self, name):
        # Breaking here? Maybe reahl refresh needed?
        try:
            return self.name_index[name]
        except KeyError:
            raise ProjectNotFound(name)

    @classmethod
    def from_file(cls, filename, workspace):
        projects = ProjectList(workspace)
        projects.read(filename)
        return projects
        
    def collect_projects(self, workspace, directories):
        for directory in directories:
            for root, dirs, files in os.walk(os.path.abspath(directory)):
                if '.reahlignore' in files:
                    ignore_list = DirectoryList.from_file(os.path.join(root, '.reahlignore'))
                    for i in ignore_list:
                        try:
                            dirs.remove(i)
                        except ValueError:
                            pass
                elif '.reahlproject' in files:
                    project = Project.from_file(self.workspace, root)
                    self.append(project, ignore_duplicates=True)
                    if not project.has_children:
                        dirs[:] = []  # This prunes the tree so it does not walk deeper in here

    def save(self, filename):
        f = io.open(filename, 'w')
        f.writelines(['%s\n' % i.relative_directory for i in self])
        f.close()

    def delete(self, filename):
        os.remove(filename)

    def read(self, filename):
        self[:] = []
        if not os.path.isfile(filename):
            return
        with io.open(filename, 'r') as f:
            project_dirs = f.read().splitlines()
        for name in project_dirs:
            full_dir = os.path.join(self.workspace.directory, name)
            if os.path.isdir(full_dir):
                self.append(Project.from_file(self.workspace, full_dir))
            else:
                logging.getLogger(__name__).warning('Skipping %s, it does not exist anymore' % name)

    def select(self, states=None, tags=None, append=False, all_=False, negated=False):
        states = states or []
        tags = tags or []
        if append:
            selection = self.workspace.selection
        else:
            selection = ProjectList(self.workspace)
    
        for i in self:
            if (not negated) == (all_ or (i.status in states) or (len(set(i.tags) & set(tags))>0)):
                selection.append(i)
        return selection

    def select_one(self, directory):
        selection = ProjectList(self.workspace)
        for i in self:
            if i.directory == directory:
                selection.append(i)
                return selection

        selection.append(Project.from_file(self.workspace, directory))
        return selection


class Workspace(object):
    """A Workspace logically contains several Projects in development. It facilitates issuing operations on sets 
    of Projects, and administers the results.

    A Workspace is a directory hierarchy on disk, the projects in it are all checked out somewhere in this hierarchy.

    The workspace provides a location where all manner of administrative information can be saved for the projects 
    it contains.
    """
    def __init__(self, directory):
        self.directory = directory
        self.startup_directory = os.getcwd()
        self.projects = ProjectList(self)
        self.selection = ProjectList(self)

    def get_selection_filename(self, name):
        return os.path.join(self.reahl_directory, '%s.selection' % name)

    @property
    def current_selection_filename(self):
        return os.path.join(self.reahl_directory, '%s.selection' % 'current')

    @property
    def current_projects_filename(self):
        return os.path.join(self.reahl_directory, 'workspace.projects')

    @property
    def build_directory(self):
        return os.path.join(self.reahl_directory, 'build')

    def save_selection(self, name):
        return self.selection.save(self.get_selection_filename(name))

    def read_selection(self, name):
        return self.selection.read(self.get_selection_filename(name))

    def delete_selection(self, name):
        return self.selection.delete(self.get_selection_filename(name))

    def save(self):
        self.selection.save(self.current_selection_filename)
        self.projects.save(self.current_projects_filename)

    def read_current_selection(self):
        self.selection.read(self.current_selection_filename)

    def read(self):
        self.read_current_selection()
        self.projects.read(self.current_projects_filename)

    def refresh(self, append, directories):
        directories = [os.path.normpath(os.path.join(self.startup_directory, i)) for i in directories]

        if not append:
            self.projects = ProjectList(self)
        self.projects.collect_projects(self, directories or [self.directory])
        self.selection = ProjectList(self)
        self.save()

    def select(self, states=None, tags=None, append=False, all_=False, negated=False):
        self.selection = self.projects.select(states=states, tags=tags, append=append, all_=all_, negated=negated)
        self.save()

    def clear_selection(self):
        self.selection = ProjectList(self)
        self.save()

    def get_selection_subset(self, states=None, tags=None, append=False, all_=False, negated=False):
        return self.selection.select(states=None, tags=None, append=False, all_=False, negated=False)

    def project_named(self, name):
        return self.projects.project_named(name)

    def has_project_named(self, name):
        return self.projects.has_project_named(name)

    def project_in(self, directory):
        return self.projects.project_in(directory)
        
    def update_apt_repository_index(self):
        self.distribution_apt_repository.build_index_files()


    @property
    def reahl_directory(self):
        dirname = os.path.join(self.directory, '.reahlworkspace')
        if not os.path.isdir(dirname):
            os.mkdir(dirname)
        return dirname

    @property
    def release_directory(self):        
        dirname = os.path.join(self.reahl_directory, 'torelease')
        if not os.path.isdir(dirname):
            os.mkdir(dirname)
        return dirname

    @property
    def distribution_apt_repository(self):
        return LocalAptRepository(os.path.join(self.reahl_directory, 'dist'))

    @property
    def distribution_egg_repository(self):
        return LocalRepository(os.path.join(self.reahl_directory, 'dist-egg'))

    @property
    def repository_state_directory(self):
        dirname = os.path.join(self.reahl_directory, 'uploadstate')
        if not os.path.isdir(dirname):
            os.mkdir(dirname)
        return dirname
        
    @property
    def dh_make_directory(self):
        dirname = os.path.join(self.reahl_directory, 'dh_make')
        if not os.path.isdir(dirname):
            None
        return dirname
        
    def get_saved_selections(self):
        filenames = glob.glob(os.path.join(self.reahl_directory, '*.selection'))
        return [os.path.splitext(os.path.basename(i))[0] for i in filenames]


class SubstvarsFile(list):
    def __init__(self, filename):
        self.filename = filename

    def read(self):
        if not os.path.isfile(self.filename):
            return

        with io.open(self.filename) as f:
            for line in f:
                bits = line.strip().split('=')
                key = bits[0]
                value = '='.join(bits[1:])
                self[key] = value

    def write(self):
        with io.open(self.filename, 'w') as f:
            for key, value in self:
                f.write('%s=%s\n' % (key, value))

    def __setitem__(self, key, value):
        try:
            [pair] = [i for i in self
                      if i[0] == key]
            self.insert(self.index(pair), (key, value))
            self.remove(pair)
        except ValueError:
            self.append((key, value))

    def __getitem__(self, key):
        return dict(self)[key]
                
    def items(self):
        return [i for i in self]




# Copyright 2013, 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the devdomain module."""

from __future__ import print_function, unicode_literals, absolute_import, division
import six
import os
import os.path
import filecmp

from nose.tools import istest
from reahl.tofu import test, Fixture
from reahl.tofu import temp_dir, temp_file_with, vassert, expected
from reahl.stubble import exempt
from reahl.stubble import stubclass

from reahl.component.shelltools import Executable
from reahl.dev.devdomain import DebianPackage, SshRepository, LocalAptRepository, RepositoryLocalState, Workspace, \
    Version, Project, ProjectMetadata, EggProject, ChickenProject, Project, SubstvarsFile, \
    Dependency, ThirdpartyDependency, DebianChangelog, DebianControl
from reahl.dev.exceptions import AlreadyUploadedException, NotBuiltException, NotAValidProjectException, \
    InvalidProjectFileException


class RepositoryUploadFixture(Fixture):
    def new_incoming_directory(self):
        return temp_dir()
    def new_to_upload_directory(self):
        return temp_dir()
    def new_state_directory(self):
        return temp_dir()

    def new_to_upload_file1(self):
        return self.to_upload_directory.file_with('piet.changes', 'kerneels')
    def new_to_upload_file2(self):
        return self.to_upload_directory.file_with('piet.deb', 'black cat')
    
    def new_package(self):
        fixture = self
        @stubclass(DebianPackage)
        class PackageStub(object):
            _built = False
            unique_id = 'therecanonlybeone'
            @property
            def is_built(self):
                return self._built

            @property
            def files_to_distribute(self):
                return [fixture.to_upload_file1.name, fixture.to_upload_file2.name]

            def build(self):
                self._built = True

        return PackageStub()

    def new_workspace(self):
        fixture = self
        @stubclass(Workspace)
        class WorkspaceStub(object):
            repository_state_directory = fixture.state_directory.name

        return WorkspaceStub()

    def new_debian_repository(self):
        user = os.environ['USER']
        return SshRepository(self.workspace, 'localhost', user, self.incoming_directory.name)


@istest
class RepositoryTests(object):
    """Tests for the SshRepository class."""
    @test(RepositoryUploadFixture)
    def upload_not_built(self, fixture):
        # Case where you upload something that's not been built yet
        vassert( not fixture.debian_repository.is_uploaded(fixture.package) )
        vassert( not fixture.package.is_built )
        with expected(NotBuiltException):
            fixture.debian_repository.upload(fixture.package, [])

    @test(RepositoryUploadFixture)
    def upload_success(self, fixture):
        fixture.package.build()

        # Case where it works
        vassert( not fixture.debian_repository.is_uploaded(fixture.package) )
        vassert( fixture.package.is_built )

        fixture.debian_repository.upload(fixture.package, [])
        vassert( fixture.debian_repository.is_uploaded(fixture.package) )
        vassert( fixture.package.files_to_distribute )
        for filename in fixture.package.files_to_distribute:
            filename_only = os.path.basename(filename)
            incoming_filename = os.path.join(fixture.incoming_directory.name, filename_only)
            vassert( filecmp.cmp(filename, incoming_filename) )

        # Case where you try upload something again
        vassert( fixture.debian_repository.is_uploaded(fixture.package) )
        vassert( fixture.package.is_built )
        with expected(AlreadyUploadedException):
            fixture.debian_repository.upload(fixture.package, [])

    @test(Fixture)
    def reading_and_writing(self, fixture):
        repository_state_dir = temp_dir()
        @stubclass(SshRepository)
        class RepositoryStub(object):
            @property
            def unique_id(self):
                return 'myid'

            repository_state_directory = repository_state_dir.name

        repository = RepositoryStub()
        expected_repository_state_file = os.path.join(repository_state_dir.name, '%s.uploaded' % repository.unique_id)

        local_state = RepositoryLocalState(repository)

        # Case: on first read, does not break if file does not exist
        vassert( not os.path.exists(expected_repository_state_file) )
        local_state.read()
        vassert( local_state.uploaded_project_ids == set([]) )

        # Case: on write, creates file
        vassert( not os.path.exists(expected_repository_state_file) )
        local_state.uploaded_project_ids = {'someid1', 'someid2'}
        local_state.write()
        vassert( os.path.isfile(expected_repository_state_file) )

        # Case: read existing stuff correctly
        local_state.uploaded_project_ids = set([])
        local_state.read()
        vassert( local_state.uploaded_project_ids == {'someid1', 'someid2'} )

    @test(Fixture)
    def queries(self, fixture):
        @stubclass(DebianPackage)
        class PackageStub(object):
            def __init__(self, name):
                self.name = name

            @property
            def unique_id(self):
                return self.name

        package1 = PackageStub('myname')
        package2 = PackageStub('yourname')


        @stubclass(SshRepository)
        class RepositoryStub(object):
            pass

        repository = RepositoryStub()
        local_state = RepositoryLocalState(repository)

        local_state.set_uploaded(package1)

        vassert( local_state.is_uploaded(package1) )
        vassert( not local_state.is_uploaded(package2) )



@stubclass(DebianPackage)
class DebianPackageStub(object):
    package_files = ['equivs-dummy_1.0_all.deb',  'equivs-dummy_1.0.dsc',
                     'equivs-dummy_1.0_i386.changes', 'equivs-dummy_1.0.tar.gz']
    def __init__(self):
        self.temp_directory = temp_dir()
        self.create_files()
        
    @exempt
    def create_files(self):
        self.file1 = self.temp_directory.file_with(self.package_files[0], 'aaa')
        self.file2 = self.temp_directory.file_with(self.package_files[1], 'bbb')
        self.file3 = self.temp_directory.file_with(self.package_files[2], 'bbb')
        self.file4 = self.temp_directory.file_with(self.package_files[3], 'bbb')
    
    @property
    def build_output_files(self):
        return [self.file1.name, self.file2.name, self.file3.name, self.file4.name]


class DebianPackageStubWithRealFiles(DebianPackageStub):
    @property
    def build_output_files(self):
        return [os.path.join(self.temp_directory.name, filename) for filename in self.package_files]

    @exempt
    def create_files(self):
        control_file_contents = """Section: misc
Priority: optional
Standards-Version: 3.6.2

Package: equivs-dummy
Version: 1.0
Maintainer: %s <%s>
Architecture: all
Description: some wise words 
 long description and info
 .
 second paragraph

""" % (os.environ['DEBFULLNAME'], os.environ['EMAIL'])
        self.temp_directory.file_with('control', control_file_contents)
        Executable('equivs-build').check_call(['-a', 'i386', '-f','control'], cwd=self.temp_directory.name)

class LocalAptRepositoryFixture(Fixture):
    def new_repository_directory(self):
        return temp_dir()

    def new_repository(self):
        return LocalAptRepository(os.path.join(self.repository_directory.name, 'repo'))
    
    def new_package(self):
        return DebianPackageStub()


@istest
class LocalAptRepositoryTests(object):
    @test(LocalAptRepositoryFixture)
    def creation_of_directory(self, fixture):
        package = fixture.package
        repository = fixture.repository
        
        # Case: when a repository is made the first time, it creates its root_directory
        vassert( os.path.isdir(repository.root_directory) )
        repository.upload(package, [])
        vassert( repository.is_uploaded(package) )

        # Case: when a repository is made a second time, it preserves the underlying directory
        repository2 = LocalAptRepository(repository.root_directory)
        vassert( repository.is_uploaded(package) )
        vassert( repository2.is_uploaded(package) )
        
    @test(LocalAptRepositoryFixture)
    def uploading(self, fixture):
        package = fixture.package
        repository = fixture.repository
        
        # Case: uploading files
        vassert( not repository.is_uploaded(package) )
        repository.upload(package, [])
        vassert( repository.is_uploaded(package) )
        vassert( package.package_files )

        for filename in package.package_files:
            vassert( os.path.isfile(os.path.join(repository.root_directory, filename)) )

        # Case: removing uploaded
        repository.remove_uploaded(package)
        for filename in repository.uploaded_files_for(package):
            vassert( not os.path.isfile(filename) )

    @test(LocalAptRepositoryFixture)
    def index_building(self, fixture):
        package = DebianPackageStubWithRealFiles()
        repository = fixture.repository
        repository.upload(package, [])
        
        repository.build_index_files()
        for filename in ['Packages','Release','Release.gpg']:
            vassert( os.path.isfile( os.path.join(repository.root_directory, filename)) ) 
        

@istest
class ProjectReadingTests(object):
    """Tests for reading Projects from XML files."""
    @test(Fixture)
    def generic_project_file_queries(self, fixture):
        workspace_dir = temp_dir()
        project_dir = workspace_dir.temp_dir()
        project_filename = os.path.join(project_dir.name, '.reahlproject')

        @stubclass(Workspace)
        class WorkspaceStub(object):
            directory = workspace_dir.name
        workspace = WorkspaceStub()

        # Case where the file does not exist
        vassert( not os.path.exists(project_filename) )
        with expected(NotAValidProjectException):
            Project.from_file(workspace, project_dir.name )

        # Case where the file exists, but is empty
        xml_file = project_dir.file_with('.reahlproject', '')
        with expected(InvalidProjectFileException):
            Project.from_file(workspace, project_dir.name)

        # Case where the file exists with stuff in it
        del xml_file
        xml_file = project_dir.file_with('.reahlproject', '''
<project type="egg">
  <distpackage type="deb">
    <sshdirectory host="localhost1" destination="/a/b"/>
    <sshdirectory host="localhost2" login="someusername" destination="/a/c"/>
  </distpackage>
</project>
''')
        project = Project.from_file(workspace, project_dir.name)

        vassert( isinstance(project, Project) )

        [package] = project.packages_to_distribute
        vassert( isinstance(package, DebianPackage) )
        vassert( package.project is project )

        [repository1, repository2] = package.repositories
        vassert( repository1.host == 'localhost1' )
        vassert( repository1.login == os.environ['USER'] )
        vassert( repository1.destination == '/a/b' )
        vassert( repository2.host == 'localhost2' )
        vassert( repository2.login == 'someusername' )
        vassert( repository2.destination == '/a/c' )

    @test(Fixture)
    def setup_project_file_queries(self, fixture):
        workspace_dir = temp_dir()
        project_dir = workspace_dir.temp_dir()
        project_filename = os.path.join(project_dir.name, '.reahlproject')
        project_source = project_dir.sub_dir('this')
        project_source_init = project_source.file_with('__init__.py', '')
        project_package1 = project_source.sub_dir('pack1')
        project_package1_init = project_package1.file_with('__init__.py', '')
        project_package2 = project_source.sub_dir('pack2')
        project_package2_init = project_package2.file_with('__init__.py', '')
        project_egg = project_dir.file_with('projegg.py', '')
        project_dev = project_source.sub_dir('proj_dev')

        @stubclass(Project)
        class ProjectStub(object):
            version = Version('1.2.5')
        
        @stubclass(Workspace)
        class WorkspaceStub(object):
            directory = workspace_dir.name
            project_name = 'proj'
            projects = []
            def project_named(self, name):
                return ProjectStub()
            def has_project_named(self, name):
                return True
        workspace = WorkspaceStub()

        # Case where the file does not exist
        vassert( not os.path.exists(project_filename) )
        with expected(NotAValidProjectException):
            Project.from_file(workspace, project_dir.name)

        # Case where the file exists, but is empty
        xml_file = project_dir.file_with('.reahlproject', '')
        with expected(InvalidProjectFileException):
            Project.from_file(workspace, project_dir.name)

        # Case where the file exists with stuff in it
        del xml_file
        xml_file = project_dir.file_with('.reahlproject', '''
<project type="egg" packagedata="included">
  <namespaces>
    <package name="this"/>
  </namespaces>

  <tag name="sometag"/>

  <distpackage type="deb">
    <sshdirectory host="localhost1" destination="/a/b"/>
  </distpackage>

  <deps purpose="run">
    <egg name="reahl-xmlreader-run"/>
  </deps>

  <deps purpose="test">
    <egg name="reahl-xmlreader-test"/>
  </deps>

  <deps purpose="build">
    <egg name="reahl-xmlreader-build"/>
  </deps>

  <script name="script1" locator="some script"/>
  <script name="script2" locator="some other script"/>
  
  <export entrypoint="entrypoint name 1" name="name1" locator="locator1"/>
  <export entrypoint="entrypoint name 2" name="name2" locator="locator2"/>

  <excludepackage name="this.pack2"/>

  <alias name="accept" command="tofu -f Piet"/>

  <pythonpath path="stuff"/>
  <pythonpath path="stuff2"/>

  <attachedfiles filetype="css">
    <file path="file/one.css"/>
    <file path="file/two.css"/>
  </attachedfiles>

  <attachedfiles filetype="js">
    <file path="file/one.js"/>
    <file path="file/two.js"/>
  </attachedfiles>

</project>
''')
        project = Project.from_file(workspace, project_dir.name)

        # Default Metadata queries that will com for setup.py:
        vassert( project.project_name == os.path.basename(project_dir.name) )
        vassert( six.text_type(project.version) == '0.0' )

        @stubclass(ProjectMetadata)
        class MetadataStub(object):
            @property
            def version(self):
                return Version('3.1.2a1-ubuntu1')
            @property
            def project_name(self):
                return 'test-proj'
            
        project.metadata = MetadataStub()
        vassert( six.text_type(project.version_for_setup()) == '3.1.2a1' )
        vassert( project.project_name == 'test-proj' )

        vassert( project.packages_for_setup() == ['this','this.pack1'] )
        vassert( project.py_modules_for_setup() == ['setup'] )
        vassert( project.include_package_data == True )
        vassert( project.namespace_packages_for_setup() == ['this'] )

        expected_value = ['reahl-xmlreader-run>=1.2,<1.3']
        actual = project.run_deps_for_setup()
        vassert( actual ==  expected_value )

        expected_value = ['reahl-xmlreader-test>=1.2,<1.3']
        actual = project.test_deps_for_setup() 
        vassert( actual ==  expected_value )

        expected_value = ['reahl-xmlreader-build>=1.2,<1.3']
        actual = project.build_deps_for_setup() 
        vassert( actual ==  expected_value )

        vassert( project.test_suite_for_setup() == 'this.proj_dev' )

        expected_value = {'console_scripts':   ['script1 = some script', 'script2 = some other script'],
                    'entrypoint name 2': ['name2 = locator2'],
                    'reahl.eggs':        ['Egg = reahl.component.eggs:ReahlEgg'], 
                    'entrypoint name 1': ['name1 = locator1'],
                    'reahl.attachments.js': ['0:file/one.js = reahl', '1:file/two.js = reahl'],
                    'reahl.attachments.css': ['0:file/one.css = reahl', '1:file/two.css = reahl']
                    }

        vassert( project.entry_points_for_setup() == expected_value )

        vassert( project.command_for_alias('accept') == ['tofu', '-f', 'Piet'] )
        vassert( project.command_for_alias('niemand') == None )

        vassert( project.python_path == ['stuff','stuff2'] )

        vassert( project.tags == ['sometag', 'component', 'toplevel'] )

    @test(Fixture)
    def egg_project_file_queries(self, fixture):
        workspace_dir = temp_dir()
        project_dir = workspace_dir.temp_dir()
        project_filename = os.path.join(project_dir.name, '.reahlproject')
        @stubclass(Workspace)
        class WorkspaceStub(object):
            directory = workspace_dir.name
        workspace = WorkspaceStub()

        # Case where the file exists with stuff in it
        xml_file = project_dir.file_with('.reahlproject', '''
<project type="egg" basket="some-basket">
  <distpackage type="deb">
    <sshdirectory host="localhost1" destination="/a/b"/>
    <sshdirectory host="localhost2" login="someusername" destination="/a/c"/>
  </distpackage>
</project>
''')
        project = Project.from_file(workspace, project_dir.name)

        vassert( isinstance(project, EggProject) )
        vassert( project.basket_name == 'some-basket' )

        [package] = project.packages_to_distribute
        vassert( isinstance(package, DebianPackage) )
        vassert( package.project is project )

        [repository1, repository2] = package.repositories
        vassert( repository1.host == 'localhost1' )
        vassert( repository1.login == os.environ['USER'] )
        vassert( repository1.destination == '/a/b' )
        vassert( repository2.host == 'localhost2' )
        vassert( repository2.login == 'someusername' )
        vassert( repository2.destination == '/a/c' )



@istest
class SubstvarTests(object):
    """Tests for manipulating debian substvar files."""
    @test(Fixture)
    def reading_and_writing(self, fixture):
        raw_file = temp_file_with('')
        substvars = SubstvarsFile(raw_file.name)
        substvars.extend( [('python:Version', 'python stuff'), ('reahl:Depends', 'lots of depends')] )

        substvars.write()

        substvars = SubstvarsFile(raw_file.name)
        vassert( len(substvars) == 0 )
        substvars.read()

        vassert( len(substvars) == 2 )
        vassert( list(substvars.items())[0] == ('python:Version', 'python stuff') )
        vassert( list(substvars.items())[1] == ('reahl:Depends', 'lots of depends') )

    @test(Fixture)
    def dict_interface(self, fixture):
        substvars = SubstvarsFile(None)
        substvars.extend( [('python:Version', 'python stuff'), ('reahl:Depends', 'lots of depends')] )

        vassert( substvars['python:Version'] == 'python stuff' )
        vassert( substvars['reahl:Depends'] == 'lots of depends' )

        substvars['python:Version'] = 'other stuff'
        vassert( substvars['python:Version'] == 'other stuff' )

        substvars['reahl:Goods'] = 'more goods'
        vassert( substvars['reahl:Goods'] == 'more goods' )

        vassert( len(substvars) == 3 )
        vassert( list(substvars.items())[0] == ('python:Version', 'other stuff') )
        vassert( list(substvars.items())[1] == ('reahl:Depends', 'lots of depends') )
        vassert( list(substvars.items())[2] == ('reahl:Goods', 'more goods') )


@istest
class DependencyTests(object):
    @test(Fixture)
    def types_of_dependencies(self, fixture):
        @stubclass(Workspace)
        class WorkspaceStub(object):
            def __init__(self, contains_project):
                self.contains_project = contains_project
            def project_named(self, name):
                return ProjectStub('proj-name', '2.1')
            def has_project_named(self, name):
                return True

        @stubclass(ChickenProject)
        class ChickenProjectStub(object):
            def contains_project_named(self, name):
                return True

        @stubclass(EggProject)
        class ProjectStub(object):
            def __init__(self, name, version, in_same_chicken=False, also_in_workspace=False):
                self.project_name = name
                self.version = Version(version)
                self.in_same_chicken = in_same_chicken
                self.workspace = WorkspaceStub(also_in_workspace)
            @property
            def chicken_project(self):
                return ChickenProjectStub() if self.in_same_chicken else None

        internal_dep = Dependency(ProjectStub('test-proj', '1.1', in_same_chicken=True), 'one')
        nonversioned_dep = Dependency(ProjectStub('test-proj', '1.1'), 'one', ignore_version=True)
        external_dep = Dependency(ProjectStub('test-proj', '1.1'), 'one')
        external_dep_in_workspace = Dependency(ProjectStub('test-proj', '1.1', also_in_workspace=True), 'one')
        thirdparty_dep = ThirdpartyDependency(ProjectStub('test-proj', '1.1'), 'one')

        # Case: for eggs
        actual = internal_dep.as_string_for_egg()
        vassert( actual == 'one>=1.1,<1.2' )

        actual = nonversioned_dep.as_string_for_egg()
        vassert( actual == 'one' )

        actual = external_dep.as_string_for_egg()
        vassert( actual == 'one>=2.1,<2.2' )

        actual = external_dep_in_workspace.as_string_for_egg()
        vassert( actual == 'one>=2.1,<2.2' )

        actual = thirdparty_dep.as_string_for_egg()
        vassert( actual == 'one' )

        # Case: for debs
        actual = internal_dep.as_string_for_deb()
        vassert( actual == 'python-one (>=1.1), python-one (<<1.2)' )

        actual = nonversioned_dep.as_string_for_deb()
        vassert( actual == 'python-one' )

        actual = external_dep.as_string_for_deb()
        vassert( actual == 'python-one (>=2.1), python-one (<<2.2)' )

        actual = external_dep_in_workspace.as_string_for_deb()
        vassert( actual == 'python-one (>=2.1), python-one (<<2.2)' )

        actual = thirdparty_dep.as_string_for_deb()
        vassert( actual == '' )

    @test(Fixture)
    def types_of_versions(self, fixture):
        @stubclass(EggProject)
        class ProjectStub(object):
            @property
            def chicken_project(self):
                return None

        
        alpha_dep = Dependency(ProjectStub(), 'one', version='1.2.5a1-1')
        normal_dep = Dependency(ProjectStub(), 'one', version='1.2.5-1')

        # Case: for eggs
        actual = alpha_dep.as_string_for_egg()
        vassert( actual == 'one>=1.2.5a1,<1.3' )

        actual = normal_dep.as_string_for_egg()
        vassert( actual == 'one>=1.2,<1.3' )

        # Case: for debs
        actual = alpha_dep.as_string_for_deb()
        vassert( actual == 'python-one (>=1.2), python-one (<<1.3)' )

        actual = normal_dep.as_string_for_deb()
        vassert( actual == 'python-one (>=1.2), python-one (<<1.3)' )



class ChangelogFixture(Fixture):
    def new_changelog_file(self):
        changlog_file = temp_file_with('''
python-reahl (2.0.0a1) unstable; urgency=low

  * Towards version 2.0.

 -- Iwan Vosloo <iwan@reahl.org>  Tue, 08 Feb 2011 12:03:44 +0000

python-reahl (0.8.0) unstable; urgency=low

  * Initial Release.

 -- Iwan Vosloo <iwan@reahl.org>  Wed, 22 Dec 2010 05:44:11 +0000
''')
        return changlog_file


@istest
class ChangelogTests(object):
    @test(ChangelogFixture)
    def parsing_name(self, fixture):
        changelog = DebianChangelog(fixture.changelog_file.name)
        vassert( changelog.package_name == 'python-reahl' )

    @test(ChangelogFixture)
    def parsing_version(self, fixture):
        changelog = DebianChangelog(fixture.changelog_file.name)
        vassert( changelog.version == '2.0.0a1' )


class DebianControlFixture(Fixture):
    def new_control_file(self):
        control_file = temp_file_with('''
Source: python-reahl
Section: reahl
Priority: optional
Maintainer: Reahl Software Services <info@reahl.org>
Build-Depends: debhelper (>= 7), python-support (>= 1.0), python (>= 2.6), python (<< 3.0)
Standards-Version: 3.8.3

Package: python-reahl-component
Architecture: all
Depends: python-setuptools, ${python:Depends}, ${reahl:Depends}
Provides: ${python:Provides}
Description: Reahl-component.
 Reahl-component - long description.

Package: python-reahl-stubble
Architecture: all
Depends: ${python:Depends}, ${reahl:Depends}
Provides: ${python:Provides}
Description: Stub tools for use in unit testing
 one line of description
 another line of description

''' )
        return control_file


@istest
class DebianControlTests(object):
    @test(DebianControlFixture)
    def descriptions(self, fixture):
        control = DebianControl(fixture.control_file.name)

        vassert( control.get_short_description_for('python-reahl-stubble') == 'Stub tools for use in unit testing' )
        expected_long = 'one line of description another line of description '
        vassert( control.get_long_description_for('python-reahl-stubble') == expected_long )

        vassert( control.get_short_description_for('python-reahl-component') == 'Reahl-component.' )

    @test(DebianControlFixture)
    def maintainer_info(self, fixture):
        control = DebianControl(fixture.control_file.name)

        vassert( control.maintainer_name == 'Reahl Software Services' )
        vassert( control.maintainer_email == 'info@reahl.org' )





# Copyright 2013, 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Copyright (C) 2006 Reahl Software Services (Pty) Ltd.  All rights reserved. (www.reahl.org)


from __future__ import print_function, unicode_literals, absolute_import, division
from reahl.tofu import Fixture, set_up, tear_down

from reahl.component.context import ExecutionContext
from reahl.component.dbutils import SystemControl
from reahl.component.config import StoredConfiguration


class CleanDatabase(Fixture):
    """A Fixture to be used as run fixture. Upon set up, it creates a new empty database with the
       correct database schema for the project and sets up any persistent classes for use with that 
       schema. It also connects to the database. Upon tear down, the Fixture disconnects from the database.
    """
    commit = False

    def new_reahlsystem(self):
        return self.config.reahlsystem

    def new_config(self):
        config = StoredConfiguration('etc/')
        config.configure()
        return config

    def new_context(self, config=None, system_control=None):
        context = ExecutionContext()
        context.set_config( config or self.config )
        context.set_system_control(system_control or self.system_control)
        return context

    def new_system_control(self):
        return SystemControl(self.config)

    def new_test_dependencies(self):
        return []

    @set_up
    def init_database(self):
        with self.context:
            orm_control = self.config.reahlsystem.orm_control
            for dependency in self.test_dependencies:
                orm_control.instrument_classes_for(dependency)
            if not self.system_control.is_in_memory:
                self.system_control.initialise_database(yes=True)
            self.system_control.connect()

    @tear_down
    def disconnect(self):
        with self.context:
            self.system_control.disconnect()







# Copyright 2013, 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The module contains code to implement commands that can be issued from a commandline to manipulate Reahl projects."""
from __future__ import print_function, unicode_literals, absolute_import, division
import six
import sys
import os
import os.path
import subprocess
from subprocess import CalledProcessError
import traceback


import shlex

from reahl.component.shelltools import Command, ReahlCommandline, Executable
from reahl.component.config import EntryPointClassList, Configuration

from reahl.dev.devdomain import Workspace, Project, ProjectList, ProjectNotFound, LocalAptRepository
from reahl.dev.exceptions import StatusException, AlreadyUploadedException, NotBuiltException, \
    NotUploadedException, NotVersionedException, NotCheckedInException, \
    MetaInformationNotAvailableException, AlreadyDebianisedException, \
    MetaInformationNotReadableException, UnchangedException, NeedsNewVersionException, \
    AlreadyMarkedAsReleasedException, NotBuiltAfterLastCommitException, NotBuiltException, \
    NotAValidProjectException
from functools import reduce


class DevShellConfig(Configuration):
    commands = EntryPointClassList('reahl.dev.commands', description='All commands (classes) that can be handled by the development shell')

    
class WorkspaceCommand(Command):
    """A specialised form of Command which is used in development. Its operations are assumed to often involve
    the development workspace.
    """

    def __init__(self, commandline):
        super(WorkspaceCommand, self).__init__(commandline)

        if 'REAHLWORKSPACE' in os.environ:
            workspace_dir = os.environ['REAHLWORKSPACE']
        else:
            workspace_dir = os.path.expanduser('~')
            print('REAHLWORKSPACE environment variable not set, defaulting to %s' % workspace_dir, file=sys.stderr)

        work_directory = os.path.abspath(os.path.expanduser(workspace_dir))
        self.workspace = Workspace(workspace_dir)
        self.workspace.read()


class Refresh(WorkspaceCommand):
    """Reconstructs the working set of projects by searching through the given directories (by default the entire workspace directory)."""
    keyword = 'refresh'
    usage_args = '[directory...]'
    options = [('-A', '--append', dict(action='store_true', dest='append', 
                                       help='append to the current working set'))]
    def execute(self, options, args):
        self.workspace.refresh(options.append, args)


class ExplainLegend(WorkspaceCommand):
    """Prints an explanation of project status indicators."""
    keyword = 'explainlegend'
    def execute(self, options, args):
        for status in StatusException.get_statusses():
            print('%s\t%s' % (status.legend, six.text_type(status())))


class Select(WorkspaceCommand):
    """Selects a subset of projects in the workspace based on their state."""
    keyword = 'select'
    options = [('-a', '--all', dict(action='store_true', dest='all', 
                                    help='operate on all projects in the workspace')),
               ('-A', '--append', dict(action='store_true', dest='append', 
                                    help='append to the current selection')),
               ('-s', '--state', dict(action='append', dest='states', default=[],
                                      help='operate on projects with given state')),
               ('-n', '--not', dict(action='store_true', dest='negated', 
                                    help='negates a state selection')),
               ('-t', '--tagged', dict(action='append', dest='tags', default=[],
                                       help='operate on projects tagged as specified'))]

    def execute(self, options, args):
        self.workspace.select(states=options.states, tags=options.tags, append=options.append, all_=options.all, negated=options.negated)


class ClearSelection(WorkspaceCommand):
    """Clears the current selection."""
    keyword = 'clearselection'

    def execute(self, options, args):
        self.workspace.clear_selection()


class ListSelections(WorkspaceCommand):
    """Lists all the named sets of selections that have been saved previously."""
    keyword = 'selections'
    def execute(self, options, args):
        for i in self.workspace.get_saved_selections():
            print(i)


class List(WorkspaceCommand):
    """Lists all the projects currently selected (or, optionally, all projects in the workspace)"""
    keyword = 'list'
    options = [('-s', '--state', dict(action='store_true', dest='with_status',
                                      help='outputs the status too')),
               ('-a', '--all', dict(action='store_true', dest='all',
                                      help='lists all, not only selection')),
               ('-d', '--directories', dict(action='store_true', dest='directories',
                                      help='lists relative directory names instead of project names'))]
    def execute(self, options, args):
        if options.all:
            project_list = self.workspace.projects
        else:
            project_list = self.workspace.selection

        for i in project_list:
            status_string = ''
            if options.with_status:
                status_string = '%s ' % i.status

            if options.directories:
                ident_string = i.relative_directory
            else:
                ident_string = i.project_name
            print('%s%s' % (status_string, ident_string))


class Save(WorkspaceCommand):
    """Saves the current selection of projects using the given name."""
    keyword = 'save'
    usage_args = '<name>'
    def execute(self, options, args):
        self.workspace.save_selection(args[0])


class DeleteSelection(WorkspaceCommand):
    """Deletes the saved selection set with the given name."""
    keyword = 'delete'
    usage_args = '<name>'
    def execute(self, options, args):
        self.workspace.delete_selection(args[0])


class Read(WorkspaceCommand):
    """Changes the current selection to the named, previously saved selection."""
    keyword = 'read'
    usage_args = '<name>'
    def execute(self, options, args):
        self.workspace.read_selection(args[0])


class ForAllWorkspaceCommand(WorkspaceCommand):
    """Some commands are executed in turn for each project in the selection. This superclass encapsulates some
       common implementation details for such commands."""

    options = [('-s', '--selection', dict(action='store_true', dest='selection', default=False,
                                          help='operate on all projects in the current selection')),
               ('-a', '--all', dict(action='store_true', dest='all', 
                                    help='operate on all projects in the workspace')),
               ('-S', '--state', dict(action='append', dest='states', default=[],
                                      help='operate on projects with given state')),
               ('-n', '--not', dict(action='store_true', dest='negated', 
                                    help='negates a state selection')),
               ('-t', '--tagged', dict(action='append', dest='tags', default=[],
                                       help='operate on projects tagged as specified')),
               ('-p', '--pause', dict(action='store_true', dest='pause',
                                      help='pause between commands')),
               ('-X', '--summary', dict(action='store_true', dest='summary',
                                      help='prints summary at end')),
               ('-d', '--delimit-output', dict(action='store_true', dest='delimit_output',
                                      help='prints starting and ending markers around the output of each project'))]
    def function(self, project, options, args):
        return None

    def verify_commandline(self, options, args):
        count = reduce(lambda a,b: a+1 if b else a, [options.selection, options.all, options.negated], 0)
        if count > 1:
            self.parser.error('Cannot use -S, -a or -n together')

        if (options.selection or options.all) and (options.states or options.tags or options.negated):
            self.parser.error('Cannot use -s or -a with -n, -S or -t')


    def execute_one(self, project, options, args):
        with project.paths_set():
            try:
                retcode = self.function(project, options, args)
            except SystemExit as ex:
                print('Script exited: %s' % ex, file=sys.stderr)
                retcode = ex.code
            except OSError as ex:
                print('Execution failed: %s' % ex, file=sys.stderr)
                retcode = ex.errno
            except (NotVersionedException, NotCheckedInException, MetaInformationNotAvailableException, AlreadyDebianisedException,
                    MetaInformationNotReadableException, UnchangedException, NeedsNewVersionException,
                    NotUploadedException, AlreadyMarkedAsReleasedException,
                    AlreadyUploadedException, NotBuiltException, 
                    NotBuiltAfterLastCommitException, NotBuiltException) as ex:
                print(six.text_type(ex), file=sys.stderr)
                retcode = None
            except CalledProcessError as ex:
                print(six.text_type(ex), file=sys.stderr)
                retcode = ex.returncode
          
        if retcode != None:
            if isinstance(retcode, six.string_types):
                print('Child was terminated with error message: %s' % retcode, file=sys.stderr)
            elif retcode < 0:
                print('Child was terminated by signal %s' % -retcode, file=sys.stderr)
            elif retcode > 0:
                print('Child returned %s' % -retcode, file=sys.stderr)

        return retcode

    def execute(self, options, args):
        if options.states or options.tags:
            project_list = self.workspace.get_selection_subset(states=options.states, tags=options.tags, append=False, all_=options.all, negated=options.negated)
        elif options.selection:
            project_list = self.workspace.selection
        else:
            project_list = ProjectList(self.workspace)
            try:
                current_project = self.workspace.project_in(os.getcwd())
            except ProjectNotFound:
                try:
                    current_project = Project.from_file(self.workspace, os.getcwd())
                except NotAValidProjectException:
                    current_project = Project(self.workspace, os.getcwd())
            project_list.append(current_project)

        pause = options.pause
        summary = options.summary
        delimit_output = options.delimit_output

        results = {}
        for i in project_list:
            if delimit_output:
                print('\n--- START %s %s ---' % (i.relative_directory, ' '.join(args)))
            results[i] = self.execute_one(i, options, args)
            if pause:
                print('--- PAUSED, hit <enter> to continue, ^D to stop ---')
                if not sys.stdin.readline():
                    print('\n^D pressed, halting immediately')
                    break
            if delimit_output:
                print('--- END %s %s ---' % (i.relative_directory, ' '.join(args)))

        if summary:
            print('\n--- SUMMARY ---')
            for i in project_list:
                print('%s %s' % (results[i], i.relative_directory), file=sys.stdout)

        self.perform_post_command_duties()

        if set(results.values()) == {0}:
            return 0
        return -1

    def perform_post_command_duties(self):
        pass


class AliasWorkspaceCommand(ForAllWorkspaceCommand):
    def __init__(self, commandline, keyword):
        super(AliasWorkspaceCommand, self).__init__(commandline)
        self.keyword = keyword
        self.failures = []

    def function(self, project, options, args):
        full_command = project.command_aliases.get(self.keyword, None)
        if not full_command:
            self.failures.append(project)
            return -1
        command = self.commandline.command_named(full_command[0])
        dash_index = full_command.index('--')
        return command.execute_one(project, options, full_command[dash_index+1:]+args) 
        

class Debianise(ForAllWorkspaceCommand):
    """Debianises a project."""
    keyword = 'debianise'
    def function(self, project, options, args):
        assert hasattr(project.metadata, 'debianise'), \
               'This command is only valid on debian-based projects (with <metadata type="debian"/>)'
        project.metadata.debianise()
        return 0


class Info(ForAllWorkspaceCommand):
    """Prints information about a project."""
    keyword = 'info'
    def print_heading(self, heading):
        print('')
        print(heading)
        print('\t'+('-'*(len(heading)+6)))

    def function(self, main_project, options, args):
        projects = [main_project]
        if main_project.has_children:
            projects = main_project.egg_projects

        for project in projects:
            self.print_heading('\tProject:\t%s' % project.directory)
            print('\tIs version controlled?:\t%s' % project.is_version_controlled())
            print('\tLast commit:\t\t%s' % project.source_control.last_commit_time)
            print('\tUnchanged?:\t\t%s' % project.is_unchanged())
            print('\tNeeds new version?:\t%s' % project.needs_new_version())
            print('\tIs checked in?:\t\t%s' % project.is_checked_in())

        self.print_heading('\tProject:\t%s' % main_project.directory)
        self.print_heading('\tPackages to distribute:')
        for package in main_project.packages_to_distribute:
            print('\t%s' % six.text_type(package))


class ForAllParsedWorkspaceCommand(ForAllWorkspaceCommand):
    """Some commands, for example an arbitrary shell command, are not directly implemented here.  Rather, they are
    passed on to some other processing unit, such as the shell.  They are in fact embedded in the command given to the
    reahl commandline processor whose only function is to make sure the same command is executed on each project
    in the selection.

    This superclass contains common implementation details for parsing such commands that are in fact embedded inside
    another commandline.  Optparse cannot directly parse such composite lines.
    """
    def __init__(self, commandline):
        super(ForAllParsedWorkspaceCommand, self).__init__(commandline)
        self.saved_args = None

    def parse_commandline(self, line):
        split_line = shlex.split(line)
        if '--' in split_line:
            separator = split_line.index('--')
            extracted_line = ' '.join(split_line[:separator])
            self.saved_args = split_line[separator+1:] # Saved for use in execute later
        else:
            extracted_line = line
            self.saved_args = []

        return super(ForAllParsedWorkspaceCommand, self).parse_commandline(extracted_line)

    def execute(self, options, args):
        return super(ForAllParsedWorkspaceCommand, self).execute(options, self.saved_args)


class Shell(ForAllParsedWorkspaceCommand):
    """Executes a shell command in each selected project, from each project's own root directory."""
    keyword = 'shell'
    usage_args = '-- <shell_command> [shell_command_options]'
    def function(self, project, options, args):
        if not args:
            print('No shell command specified to run', file=sys.stderr)
            return 1
        replaced_command = []
        for i in args:
            if i.startswith('$(') and i.endswith(')'):
                shellcommand = i[2]
                shell_args = i[3:-1].split(' ')
                output = Executable(shellcommand).Popen(shell_args, cwd=project.directory, stdout=subprocess.PIPE).communicate()[0]
                for line in output.splitlines():
                    replaced_command.append(line)
            else:
                replaced_command.append(i)

        command = replaced_command

        return Executable(command[0]).call(command[1:], cwd=project.directory)



class Setup(ForAllParsedWorkspaceCommand):
    """Runs setup.py <command> for each project in the current selection."""
    keyword = 'setup'
    usage_args = '-- <setup_command> [setup_command_options]'
    def function(self, project, options, args):
        project.setup(args)
        return 0


class Build(ForAllWorkspaceCommand):
    """Builds all distributable packages for each project in the current selection."""
    keyword = 'build'
    def function(self, project, options, args):
        try:
            return_code = project.build()
        except:
            print('', file=sys.stderr)
            traceback.print_exc()
            return_code = -1

        return return_code

    def perform_post_command_duties(self):
        self.workspace.update_apt_repository_index()


class ListMissingDependencies(ForAllWorkspaceCommand):
    """Lists all missing thirdparty dependencies for each project in the current selection."""
    keyword = 'missingdeps'
    options = ForAllWorkspaceCommand.options+[('-D', '--development', dict(action='store_true', dest='for_development', default=False,
                                                                           help='include development dependencies'))]
    def function(self, project, options, args):
        return_code = 0
        try:
            dependencies = project.list_missing_dependencies(for_development=options.for_development)
            if dependencies:
                print(' '.join(dependencies))
        except:
            traceback.print_exc()
            return_code = -1

        return return_code


class DebInstall(ForAllParsedWorkspaceCommand):
    """Runs setup.py install with correct arguments for packaging the result in a deb (for each project in the selection)."""
    keyword = 'debinstall'
    def function(self, project, options, args):
        project.debinstall(args)
        return 0


class Upload(ForAllWorkspaceCommand):
    """Uploads all built distributable packages for each project in the current selection."""
    keyword = 'upload'
    options = ForAllWorkspaceCommand.options+[('-k', '--knock', dict(action='append', dest='knocks', default=[],
                                                                     help='port to knock on before uploading'))]
    def function(self, project, options, args):
        return project.upload(knocks=options.knocks)


class MarkReleased(ForAllWorkspaceCommand):
    """Marks each project in the current selection as released."""
    keyword = 'markreleased'
    def function(self, project, options, args):
        return project.mark_as_released()


class SubstVars(ForAllWorkspaceCommand):
    """Generates debian substvars."""
    keyword = 'substvars'
    def function(self, project, options, args):
        assert hasattr(project.metadata, 'generate_deb_substvars'), \
               'This command is only valid on debian-based projects (with <metadata type="debian"/>)'
        project.metadata.generate_deb_substvars()
        return 0


class UpdateAptRepository(WorkspaceCommand):
    """Updates the index files of the given apt repository."""
    keyword = 'updateapt'
    usage_args = '<root_directory>'
    def execute(self, options, args):
        root_directory = None
        if args:
            root_directory = args[0]
        else:
            raise Exception('No root_directory specified')

        if not os.path.isdir( root_directory ):
            message = '"%s" is not a valid directory from within %s' % (root_directory, os.getcwd())
            raise Exception(message)

        repository = LocalAptRepository( root_directory )
        repository.build_index_files()


class ExtractMessages(ForAllWorkspaceCommand):
    """Collects strings marked for translation."""
    keyword = 'extractmessages'
    def function(self, project, options, args):
        project.extract_messages(args)
        return 0


class MergeTranslations(ForAllWorkspaceCommand):
    """Merges newly discovered strings needing translation with a catalog of existing translations."""
    keyword = 'mergetranslations'
    def function(self, project, options, args):
        project.merge_translations()
        return 0


class CompileTranslations(ForAllWorkspaceCommand):
    """Compiles all current translations."""
    keyword = 'compiletranslations'
    def function(self, project, options, args):
        project.compile_translations()
        return 0


class AddLocale(ForAllWorkspaceCommand):
    """Adds a new locale catalog."""
    keyword = 'addlocale'
    usage_args = '<locale_string> [translated_egg]'
    def function(self, project, options, args):
        if len(args) < 1:
            self.parser.error('At least a locale_string should be given')

        locale = args[0]
        translated_egg = locale
        if len(args) >= 2:
            translated_egg = args[1]

        project.add_locale(locale, translated_egg)
        return 0


class WorkspaceCommandline(ReahlCommandline):
    """The main class for invoking commands on the commandline of a development Workspace."""
    usage_string = '[options] <command> [command options] [-- [command_argument...]]'
    def __init__(self, options):
        super(WorkspaceCommandline, self).__init__(options, DevShellConfig())

    def execute_command(self, command, line, options, parser):
        self.set_log_level(options.loglevel)
        if command in self.command_names:
            return self.command_named(command).do(line)

        alias_command = AliasWorkspaceCommand(self, command)
        retcode = alias_command.do(line)
        if alias_command.failures:
            print('\nThere is no command named %s\nAlso, no aliases found for it in project(s):\n' % command, file=sys.stderr)
            print(', '.join([i.project_name for i in alias_command.failures]), file=sys.stderr)
            self.print_usage(parser)
            retcode = -1
        return retcode







from setuptools import setup
setup(
    name=u'reahl-dev',
    version='3.0.0',
    description=u'The core Reahl development tools.',
    long_description=u'Reahl is a web application framework that allows a Python programmer to work in terms of useful abstractions - using a single programming language.\n\nReahl-dev is the component containing general Reahl development tools.\n\nSee http://www.reahl.org/docs/3.0/tutorial/gettingstarted.d.html for installation instructions. ',
    url=u'http://www.reahl.org',
    maintainer='Iwan Vosloo',
    maintainer_email='iwan@reahl.org',
    packages=['reahl', 'reahl.dev_dev', 'reahl.dev'],
    py_modules=[u'setup'],
    include_package_data=False,
    package_data={'': ['*/LC_MESSAGES/*.mo']},
    namespace_packages=['reahl'],
    install_requires=['reahl-component>=3.0,<3.1', 'reahl-tofu>=3.0,<3.1', 'reahl-bzrsupport>=3.0,<3.1', 'Babel>=1.3,<1.3.9999'],
    setup_requires=[],
    tests_require=['reahl-tofu>=3.0,<3.1', 'reahl-stubble>=3.0,<3.1'],
    test_suite=u'reahl.dev_dev',
    entry_points={
        u'reahl.dev.xmlclasses': [
            u'MetaInfo = reahl.dev.devdomain:MetaInfo',
            u'HardcodedMetadata = reahl.dev.devdomain:HardcodedMetadata',
            u'DebianPackageMetadata = reahl.dev.devdomain:DebianPackageMetadata',
            u'BzrSourceControl = reahl.dev.devdomain:BzrSourceControl',
            u'Project = reahl.dev.devdomain:Project',
            u'ChickenProject = reahl.dev.devdomain:ChickenProject',
            u'EggProject = reahl.dev.devdomain:EggProject',
            u'DebianPackage = reahl.dev.devdomain:DebianPackage',
            u'DebianPackageSet = reahl.dev.devdomain:DebianPackageSet',
            u'SshRepository = reahl.dev.devdomain:SshRepository',
            u'PythonSourcePackage = reahl.dev.devdomain:PythonSourcePackage',
            u'PackageIndex = reahl.dev.devdomain:PackageIndex',
            u'Dependency = reahl.dev.devdomain:Dependency',
            u'ThirdpartyDependency = reahl.dev.devdomain:ThirdpartyDependency',
            u'XMLDependencyList = reahl.dev.devdomain:XMLDependencyList',
            u'ExtrasList = reahl.dev.devdomain:ExtrasList',
            u'EntryPointExport = reahl.dev.devdomain:EntryPointExport',
            u'ScriptExport = reahl.dev.devdomain:ScriptExport',
            u'NamespaceList = reahl.dev.devdomain:NamespaceList',
            u'NamespaceEntry = reahl.dev.devdomain:NamespaceEntry',
            u'PersistedClassesList = reahl.dev.devdomain:PersistedClassesList',
            u'OrderedPersistedClass = reahl.dev.devdomain:OrderedPersistedClass',
            u'FileList = reahl.dev.devdomain:FileList',
            u'AttachmentList = reahl.dev.devdomain:AttachmentList',
            u'ShippedFile = reahl.dev.devdomain:ShippedFile',
            u'MigrationList = reahl.dev.devdomain:MigrationList',
            u'ConfigurationSpec = reahl.dev.devdomain:ConfigurationSpec',
            u'ScheduledJobSpec = reahl.dev.devdomain:ScheduledJobSpec',
            u'ExcludedPackage = reahl.dev.devdomain:ExcludedPackage',
            u'TranslationPackage = reahl.dev.devdomain:TranslationPackage',
            u'CommandAlias = reahl.dev.devdomain:CommandAlias',
            u'ExtraPath = reahl.dev.devdomain:ExtraPath',
            u'ProjectTag = reahl.dev.devdomain:ProjectTag'    ],
        u'console_scripts': [
            u'reahl = reahl.dev.devshell:WorkspaceCommandline.execute_one'    ],
        u'reahl.eggs': [
            u'Egg = reahl.component.eggs:ReahlEgg'    ],
        u'reahl.dev.commands': [
            u'Refresh = reahl.dev.devshell:Refresh',
            u'ExplainLegend = reahl.dev.devshell:ExplainLegend',
            u'List = reahl.dev.devshell:List',
            u'Select = reahl.dev.devshell:Select',
            u'ClearSelection = reahl.dev.devshell:ClearSelection',
            u'ListSelections = reahl.dev.devshell:ListSelections',
            u'Save = reahl.dev.devshell:Save',
            u'Read = reahl.dev.devshell:Read',
            u'DeleteSelection = reahl.dev.devshell:DeleteSelection',
            u'Shell = reahl.dev.devshell:Shell',
            u'Setup = reahl.dev.devshell:Setup',
            u'Build = reahl.dev.devshell:Build',
            u'ListMissingDependencies = reahl.dev.devshell:ListMissingDependencies',
            u'DebInstall = reahl.dev.devshell:DebInstall',
            u'Upload = reahl.dev.devshell:Upload',
            u'MarkReleased = reahl.dev.devshell:MarkReleased',
            u'SubstVars = reahl.dev.devshell:SubstVars',
            u'Debianise = reahl.dev.devshell:Debianise',
            u'Info = reahl.dev.devshell:Info',
            u'ExtractMessages = reahl.dev.devshell:ExtractMessages',
            u'MergeTranslations = reahl.dev.devshell:MergeTranslations',
            u'CompileTranslations = reahl.dev.devshell:CompileTranslations',
            u'AddLocale = reahl.dev.devshell:AddLocale',
            u'UpdateAptRepository = reahl.dev.devshell:UpdateAptRepository',
            u'ServeSMTP = reahl.dev.mailtest:ServeSMTP'    ],
                 },
    extras_require={}
)



# Copyright 2013, 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

# -*- mode: python; mode: font-lock -*-
# Copyright 2005 Iwan Vosloo

from __future__ import print_function, unicode_literals, absolute_import, division
import six
import time

from reahl.stubble import easter_egg, stubclass, EmptyStub
from nose.tools import istest
from reahl.tofu import Fixture, test, set_up, tear_down
from reahl.tofu import temp_file_with, check_limitation, vassert, expected

from reahl.dev.xmlreader import XMLReader, DoubleRegistrationException, TagNotRegisteredException


#--------------------------------------------------[ fixtures ]


class BasicObjectSetup(Fixture):
    tag_name = 'anObject'
    string = 'something'
    integer = 234
    def new_file_contents(self):
        return '<%s attr1="%s" attr2="%s"/>' % (self.tag_name, self.string, self.integer)
    def new_file(self):
        return temp_file_with(self.file_contents)
    def new_test_class(self):
        class TestClass(object):
            def inflate_attributes(self, reader, attributes, parent):
                self.attr2 = int(attributes['attr2'])
                self.attr1 = six.text_type(attributes['attr1'])
                self.parent = parent
        return TestClass

    @set_up
    def init_easter_egg(self):
        easter_egg.add_to_working_set()

    def new_reader(self):
        return XMLReader()

    @tear_down
    def close_file(self):
        self.file.close()


class ConfiguredReader(BasicObjectSetup):
    class TestClass1(object):
        @classmethod
        def get_xml_registration_info(cls):
            return ('1', cls, None)

    class TestClass2(object):
        @classmethod
        def get_xml_registration_info(cls):
            return ('2', cls, 'asd')

    def new_reader(self):
        easter_egg.add_entry_point_from_line('reahl.dev.xmlclasses',
                                     'test1 = reahl.dev_dev.xmlreadertests:ConfiguredReader.TestClass1')
        easter_egg.add_entry_point_from_line('reahl.dev.xmlclasses',
                                     'test2 = reahl.dev_dev.xmlreadertests:ConfiguredReader.TestClass2')

        @stubclass(XMLReader)
        class XMLReaderStub(XMLReader):
            registered = {}
            def register(self, tag_name, class_to_instantiate, type_name=None):
                self.registered[class_to_instantiate] = (tag_name, class_to_instantiate, type_name)

        return XMLReaderStub()


class TextObjectSetup(BasicObjectSetup):
    text = 'somet ext'

    def new_test_class(self):
        class TestClass(object):
            def inflate_text(self, reader, text, parent):
                self.text = text
        return TestClass

    def new_file_contents(self):
        return '<%s>%s</%s>' % (self.tag_name, self.text, self.tag_name)


class EmptyTextObjectSetup(TextObjectSetup):
    text = ''


class CompositeObjectSetup(BasicObjectSetup):
    inner_tag_name = 'inner'
    def new_inner_class(self):
        return super(CompositeObjectSetup, self).new_test_class()
    
    def new_test_class(self):
        class OuterClass(object):
            def inflate_attributes(self, reader, attributes, parent):
                self.inner = []
                self.parent = parent
            def inflate_child(self, reader, child, tag, parent):
                self.inner.append((tag, child))

        return OuterClass

    def new_file_contents(self):
        contents = '''
        <%s>
          <%s attr1="%s" attr2="%s"/>
        </%s>
        ''' % (self.tag_name, self.inner_tag_name, self.string, self.integer, self.tag_name)
        return contents


class MultipleClassesForTag(CompositeObjectSetup):
    inner_tag_name = 'sametag'
    innerType1 = 'onetype'
    innerType2 = 'twotype'

    def new_inner_class2(self):
        class Other(object):
            pass
        return Other

    def new_file_contents(self):
        contents = '''
        <%s>
          <%s type="%s" attr1="%s" attr2="%s"/>
          <%s type="%s"/>
        </%s>
        ''' % (self.tag_name,
               self.inner_tag_name, self.innerType1, self.string, self.integer,
               self.inner_tag_name, self.innerType2,
               self.tag_name)
        return contents


class TimeStampedInflationSetup(CompositeObjectSetup):
    def new_test_class(self):
        class OuterClass(super(TimeStampedInflationSetup, self).new_test_class()):
            def inflate_attributes(self, reader, attributes, parent):
                super(OuterClass, self).inflate_attributes(reader, attributes, parent)
                self.attribute_stamp = time.time()
            def inflate_child(self, reader, child, tag, parent):
                super(OuterClass, self).inflate_child(reader, child, tag, parent)
                self.child_stamp = time.time()
            def inflate_text(self, reader, text, parent):
                self.child_stamp = time.time()
            def perform_post_inflation_checks(self):
                self.check_stamp = time.time()

        return OuterClass


#--------------------------------------------------[ tests ]
@istest
class XMLReaderTests(object):

    @test(Fixture)
    def usage_of_dom_implementation(self, fixture):
        dom_msg = 'DOM Level 3 adds a Load/Store specification, which defines an interface to the reader, but this is not yet available in the Python standard library.'
        check_limitation('2.7', 'More standardised DOM stuff may be available - this implementation uses minidom (from current docs: "%s")' % dom_msg)


    #--------------------------------------------------[ a basic object ]
    @test(BasicObjectSetup)
    def read_basic_object(self, fixture):
        fixture.reader.register(fixture.tag_name, fixture.test_class)
        parent = EmptyStub()
        read_object = fixture.reader.read_file(fixture.file, parent)

        vassert( isinstance(read_object, fixture.test_class) )
        vassert( read_object.attr1 == fixture.string )
        vassert( read_object.attr2 == fixture.integer )
        vassert( read_object.parent is parent )

    @test(BasicObjectSetup)
    def exception_on_double_register(self, fixture):
        fixture.reader.register(fixture.tag_name, fixture.test_class)

        with expected(DoubleRegistrationException):
            fixture.reader.register(fixture.tag_name, fixture.test_class)

    @test(BasicObjectSetup)
    def exception_on_not_registered(self, fixture):
        with expected(TagNotRegisteredException):
            fixture.reader.read_file(fixture.file, None)

    #--------------------------------------------------[ a composite object ]
    @test(CompositeObjectSetup)
    def read_composite_object(self, fixture):
        fixture.reader.register(fixture.inner_tag_name, fixture.inner_class)
        fixture.reader.register(fixture.tag_name, fixture.test_class)

        read_object = fixture.reader.read_file(fixture.file, None)

        vassert(isinstance(read_object, fixture.test_class))
        vassert(len(read_object.inner) == 1)
        child = read_object.inner[0][1]
        child_tag = read_object.inner[0][0]
        vassert(isinstance(child, fixture.inner_class))
        vassert(child_tag == fixture.inner_tag_name)
        vassert(child.attr1 == fixture.string)
        vassert(child.attr2 == fixture.integer)
        vassert(child.parent is read_object)

    @test(MultipleClassesForTag)
    def read_objects_with_same_tag_and_different_type_attribute(self, fixture):
        fixture.reader.register(fixture.inner_tag_name, fixture.inner_class, fixture.innerType1)
        fixture.reader.register(fixture.inner_tag_name, fixture.inner_class2, fixture.innerType2)
        fixture.reader.register(fixture.tag_name, fixture.test_class)

        read_object = fixture.reader.read_file(fixture.file, None)

        vassert(isinstance(read_object, fixture.test_class))
        vassert(len(read_object.inner) == 2)

        vassert(isinstance(read_object.inner[0][1], fixture.inner_class))
        vassert(read_object.inner[0][1].attr1 == fixture.string)
        vassert(read_object.inner[0][1].attr2 == fixture.integer)

        vassert(isinstance(read_object.inner[1][1], fixture.inner_class2))

    #--------------------------------------------------[ simple object with text content ]
    def read_text_object(self, fixture):
        fixture.reader.register(fixture.tag_name, fixture.test_class)
        read_object = fixture.reader.read_file(fixture.file, None)
        if fixture.text:
            vassert(read_object.text == fixture.text)
        else:
            vassert(not hasattr(read_object, 'text'))

    @test(TextObjectSetup)
    def read_non_empty_text_object(self, fixture):
        self.read_text_object(fixture)

    @test(EmptyTextObjectSetup)
    def read_empty_text_object(self, fixture):
        self.read_text_object(fixture)


    #--------------------------------------------------[ order of inflation method calls ]
    @test(TimeStampedInflationSetup)
    def order_of_methods_calls_is_correct(self, fixture):
        fixture.reader.register(fixture.inner_tag_name, fixture.inner_class)
        fixture.reader.register(fixture.tag_name, fixture.test_class)

        read_object = fixture.reader.read_file(fixture.file, None)

        vassert(read_object.attribute_stamp < read_object.child_stamp < read_object.check_stamp)


    #--------------------------------------------------[ configuration ]
    @test(ConfiguredReader)
    def available_classes_from_plugins_are_registered(self, fixture):
        for cls in [fixture.TestClass1, fixture.TestClass2]:
            vassert( fixture.reader.registered[cls] == cls.get_xml_registration_info() )



# Copyright 2013, 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Exceptions used by modules in this package."""

from __future__ import print_function, unicode_literals, absolute_import, division
import inspect

class StatusException(Exception):
    legend = None
    @classmethod
    def get_statusses(cls):
        import reahl.dev.exceptions
        return [i for i in [getattr(reahl.dev.exceptions, name) for name in dir(reahl.dev.exceptions)]
                if inspect.isclass(i) and issubclass(i, cls) and i is not cls]
            

class NoException(StatusException):
    legend = '+'
    def __str__(self):
        return 'New release uploaded - needs to be marked as released'

    
class NotBuiltException(StatusException):
    legend = 'B'
    def __str__(self):
        return 'Project not built'


class NotBuiltAfterLastCommitException(StatusException):
    legend = 'b'
    def __str__(self):
        return 'Project build is older than the last commit'


class NotUploadedException(StatusException):
    legend = 'U'
    def __str__(self):
        project_name = 'Project'
        if self.args:
            project_name = self.args[0]
        return '%s is not uploaded' % project_name


class NotVersionedException(StatusException):
    legend = '.'
    def __str__(self):
        return 'Not versioned'


class NotCheckedInException(StatusException):
    legend = 'C'
    def __str__(self):
        return 'Not checked in %s' % (self.args or '')


class MetaInformationNotAvailableException(StatusException):
    legend = 'M'
    def __str__(self):
        return 'Cannot find project meta information'


class MetaInformationNotReadableException(StatusException):
    legend = 'm'
    def __str__(self):
        return 'Cannot reliably read project meta information - perhaps a malformed file?'


class UnchangedException(StatusException):
    legend = '='
    def __str__(self):
        return 'Unchanged since last release'


class NeedsNewVersionException(StatusException):
    legend = 'V'
    def __str__(self):
        return 'Needs a new version'

    
class AlreadyMarkedAsReleasedException(Exception):
    def __str__(self):
        return 'Already marked as released'


class AlreadyUploadedException(Exception):
    def __str__(self):
        return 'Already uploaded'


class AlreadyDebianisedException(Exception):
    def __str__(self):
        return 'Already debianised'
        

class NotAValidProjectException(Exception):
    def __str__(self):
        return 'Not a valid project'


class InvalidProjectFileException(Exception):
    def __str__(self):
        return 'Invalid project file: %s' % self.args[0]


class NoSuchProjectException(Exception):
    def __str__(self):
        return 'There is no such project (%s) in the workspace' % self.args[0]


    



# Copyright 2013, 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""A generic factory that can construct Python classes from XML.

This package provides infrastructure that will read an XML file with
clean, human readable XML and provide you with a bunch of classes that
have been constructed and initialised according to the information in
the XML file.

The XML can look like one of the following:

 1. <mytag>asdasd</mytag>
 1. <mytag type='atype'>asfasfd</mytag>

Classes are registered for a specific XML element name and type
combination.

To be able to register a class, it should have the following class
method:

    @classmethod
    def get_xml_registration_info(cls):
        return ( <tag>, cls, <type> )

Where: cls is the class that will be instantiated when XML element
    <tag> with type attribute <type> is encountered.

    Speficying None for <type> means that cls will be instantiated iff
    an XML element with name <tag> is encountered and it does not have
    a 'type=' attribute.


Registered classes further may have the following methods which will
be called by this package in order to initialise a newly created
instance:

  def inflate_attributes(self, reader, attributes, parent)

    This is the very first call to a newly created instance.  The
    class should be initialised here and all its simple attributes
    set.

      - reader is the XMLReader which is constructing the class.
      - attributes is a dictionary with all the XML attributes
         specified for the XML element that triggered this class
         instantiation
      - parent is the object which was constructed from the parent
         element of the current XML element (it may be None)

  def inflate_child(self, reader, child, tag, parent)

    If an element has children elements, that means the constructed
    class has attributes that are also class instances.  In this case,
    as each child is inflated, this method will be called on its
    parent so that the parent can set its attribute.

      - reader is the XMLReader which is constructing the class.
      - child is the newly instantiated child class instance
      - tag is the XML element name which triggered the instantiation
        of the child
      - parent is the object which was constructed from the parent
         element of the current XML element (it may be None)

  def inflate_text(self, reader, text, parent)

    Sometimes an element will have nothing but text in it.  In such a
    case, this method is called on the constructed class

      - reader is the XMLReader which is constructing the class.
      - text is the text
      - parent is the object which was constructed from the parent
         element of the current XML element (it may be None)

  def perform_post_inflation_checks(self)

    This method is called after all other methods have been called.
    It can be used to check that a class has been properly
    initialised.


"""
from __future__ import print_function, unicode_literals, absolute_import, division

from xml.dom.minidom import parse

from reahl.component.config import Configuration, EntryPointClassList


class XMLReaderConfig(Configuration):
    xmlclasses = EntryPointClassList('reahl.dev.xmlclasses', description='All classes that can be read by the xmlreader')

#--------------------------------------------------[ XMLReaderException ]
class XMLReaderException(Exception):
    """All exceptions of this module are derived from this one."""


#--------------------------------------------------[ DoubleRegistrationException ]
class DoubleRegistrationException(XMLReaderException):
    """Raised when an attempt is made to register the same tag more than once."""


#--------------------------------------------------[ TagNotRegisteredException ]
class TagNotRegisteredException(XMLReaderException):
    """Raised when a tag was encountered during XML parsing for which no class is registered."""


#--------------------------------------------------[ TagTypeTuple ]
class TagTypeTuple(tuple):
    """A tuple of (tag,type) with easy accessors for its elements."""
    def __new__(cls, arg):
        assert len(arg) == 2, 'Programming error: you must supply both tag and type'
        return super(TagTypeTuple, cls).__new__(cls, arg)
    def tag(self):
        return self[0]
    def type(self):
        return self[1]


#--------------------------------------------------[ XMLReader ]
class XMLReader(object):
    """This class is used to read python objects from an XML file."""
    def __init__(self):
        self.map = {}
        self.register_classes_from_plugins()

    def register_classes_from_plugins(self):
        config = XMLReaderConfig()
        for cls in config.xmlclasses:
            self.register(*cls.get_xml_registration_info())

    def register(self, tag_name, class_to_instantiate, type_name=None):
        tag_type_tuple = TagTypeTuple((tag_name, type_name))

        if tag_type_tuple in self.map:
            message = 'Tag "%s" with type "%s" already registered for class %s' % \
                      (tag_name, type_name, self.map[tag_type_tuple],)
            raise DoubleRegistrationException(message)

        self.map[tag_type_tuple] = class_to_instantiate

    def read_file(self, input_file, parent):
        dom = parse(input_file)
        element = dom.documentElement

        result = self.to_object(element, parent)
        dom.unlink()

        return result

    def to_object(self, dom_element, parent_object):
        type_name = None
        if dom_element.hasAttribute('type'):
            type_name = dom_element.getAttribute('type')

        tag_type_tuple = TagTypeTuple((dom_element.tagName, type_name))

        if tag_type_tuple not in self.map:
            message = 'No class registered for tag "%s", type "%s"' % tag_type_tuple
            raise TagNotRegisteredException(message)

        class_to_instantiate = self.map[tag_type_tuple]
        instance = class_to_instantiate.__new__(class_to_instantiate)

        self.call_if_present(instance, 'inflate_attributes',
                           self, dict(dom_element.attributes.items()), parent_object)

        dom_element.normalize()
        for element in dom_element.childNodes:
            if element.nodeType == element.ELEMENT_NODE:
                child = self.to_object(element, instance)
                self.call_if_present(instance, 'inflate_child',
                                   self, child, element.tagName, parent_object)
            if element.nodeType == element.TEXT_NODE:
                text = element.data
                self.call_if_present(instance, 'inflate_text', self, text, parent_object)

        self.call_if_present(instance, 'perform_post_inflation_checks')

        return instance

    def call_if_present(self, instance, function_name, *args, **kwargs):
        if hasattr(instance, function_name):
            getattr(instance, function_name)(*args, **kwargs)


